from typing import Optional

import torch

from .potential import Potential


class CombinedPotential(Potential):
    """
    A potential that is a linear combination of multiple potentials.

    A class representing a combined potential that aggregates multiple individual
    potentials with weights for use in long-range (LR) and short-range (SR)
    interactions.

    The ``CombinedPotential`` class allows for flexible combination of potential
    functions with user-specified weights, which can be either fixed or trainable.

    :param potentials: List of potential objects, each implementing a compatible
        interface with methods `from_dist`, `lr_from_dist`, `lr_from_k_sq`,
        `self_contribution`, and `background_correction`.
    :param initial_weights: Initial weights for combining the potentials. If provided,
        the length must match the number of potentials. If `None`, weights are
        initialized to ones.
    :param learnable_weights: If `True`, weights are trainable parameters, allowing
        optimization during training. If `False`, weights are fixed.
    :param exclusion_radius: A length scale that defines a *local environment* within
        which the potential should be smoothly zeroed out, as it will be described by a
        separate model.
    :param exclusion_degree: Controls the sharpness of the transition in the cutoff function
        applied within the ``exclusion_radius``. The cutoff is computed as a raised cosine
        with exponent ``exclusion_degree``
    """

    def __init__(
        self,
        potentials: list[Potential],
        initial_weights: Optional[torch.Tensor] = None,
        learnable_weights: Optional[bool] = True,
        smearing: Optional[float] = None,
        exclusion_radius: Optional[float] = None,
        exclusion_degree: int = 1,
    ):
        super().__init__(
            smearing=smearing,
            exclusion_radius=exclusion_radius,
            exclusion_degree=exclusion_degree,
        )

        smearings = [pot.smearing for pot in potentials]
        if not all(smearings) and any(smearings):
            raise ValueError(
                r"Cannot combine direct (`smearing=None`) and range-separated (`smearing=float`) potentials."
            )

        if all(smearings) and not self.smearing:
            # this is very misleading, but it is the way the original code works,
            # otherwise mypy complains
            raise ValueError(
                r"You should specify a `smearing` when combining range-separated (`smearing=float`) potentials."
            )
        if not any(smearings) and self.smearing:
            # this is very misleading, but it is the way the original code works,
            # otherwise mypy complai
            raise ValueError(
                r"Cannot specify `smearing` when combining direct (`smearing=None`) potentials."
            )

        if initial_weights is not None:
            if len(initial_weights) != len(potentials):
                raise ValueError(
                    "The number of initial weights must match the number of potentials being combined"
                )
        else:
            initial_weights = torch.ones(len(potentials))
        # for torchscript
        self.potentials = torch.nn.ModuleList(potentials)
        if learnable_weights:
            self.weights = torch.nn.Parameter(initial_weights)
        else:
            self.register_buffer("weights", initial_weights)

    def from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        potentials = [pot.from_dist(dist) for pot in self.potentials]
        potentials = torch.stack(potentials, dim=-1)
        return torch.inner(self.weights, potentials)

    def sr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        potentials = [pot.sr_from_dist(dist) for pot in self.potentials]
        potentials = torch.stack(potentials, dim=-1)
        return torch.inner(self.weights, potentials)

    def lr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        potentials = [pot.lr_from_dist(dist) for pot in self.potentials]
        potentials = torch.stack(potentials, dim=-1)
        return torch.inner(self.weights, potentials)

    def lr_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        potentials = [pot.lr_from_k_sq(k_sq) for pot in self.potentials]
        potentials = torch.stack(potentials, dim=-1)
        return torch.inner(self.weights, potentials)

    def self_contribution(self) -> torch.Tensor:
        # self-correction for 1/r^p potential
        potentials = [pot.self_contribution() for pot in self.potentials]
        potentials = torch.stack(potentials, dim=-1)
        return torch.inner(self.weights, potentials)

    def background_correction(self) -> torch.Tensor:
        # "charge neutrality" correction for 1/r^p potential
        potentials = [pot.background_correction() for pot in self.potentials]
        potentials = torch.stack(potentials, dim=-1)
        return torch.inner(self.weights, potentials)

    from_dist.__doc__ = Potential.from_dist.__doc__
    sr_from_dist.__doc__ = Potential.sr_from_dist.__doc__
    lr_from_dist.__doc__ = Potential.lr_from_dist.__doc__
    lr_from_k_sq.__doc__ = Potential.lr_from_k_sq.__doc__
    self_contribution.__doc__ = Potential.self_contribution.__doc__
    background_correction.__doc__ = Potential.background_correction.__doc__



import torch


def get_ns_mesh(cell: torch.Tensor, mesh_spacing: float):
    """
    Computes the mesh size given a target mesh spacing and cell
    getting the closest powers of 2 to help with FFT.

    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param mesh_spacing: float
    :param differentiable: boll

    :return: torch.tensor of length 3 containing the mesh size
    """
    basis_norms = torch.linalg.norm(cell, dim=1)
    ns_approx = basis_norms / mesh_spacing
    ns_actual_approx = 2 * ns_approx + 1  # actual number of mesh points
    # ns = [nx, ny, nz], closest power of 2 (helps for FT efficiency)
    return torch.tensor(2).pow(torch.ceil(torch.log2(ns_actual_approx)).long())


def _generate_kvectors(
    cell: torch.Tensor, ns: torch.Tensor, for_ewald: bool
) -> torch.Tensor:
    # Check that all provided parameters have the correct shapes and are consistent
    # with each other
    if cell.shape != (3, 3):
        raise ValueError(f"cell of shape {list(cell.shape)} should be of shape (3, 3)")

    if ns.shape != (3,):
        raise ValueError(f"ns of shape {list(ns.shape)} should be of shape (3, )")

    if ns.device != cell.device:
        raise ValueError(
            f"`ns` and `cell` are not on the same device, got {ns.device} and "
            f"{cell.device}."
        )

    if cell.is_cuda:
        # use function that does not synchronize with the CPU
        inverse_cell = torch.linalg.inv_ex(cell)[0]
    else:
        inverse_cell = torch.linalg.inv(cell)

    reciprocal_cell = 2 * torch.pi * inverse_cell.T
    bx = reciprocal_cell[0]
    by = reciprocal_cell[1]
    bz = reciprocal_cell[2]

    # Generate all reciprocal space vectors from real FFT!
    # The frequencies from the fftfreq function  are of the form [0, 1/n, 2/n, ...]
    # These are then converted to [0, 1, 2, ...] by multiplying with n.
    # get the frequencies, multiply with n, then w/ the reciprocal space vectors
    kxs = (bx * ns[0]) * torch.fft.fftfreq(
        ns[0], device=cell.device, dtype=cell.dtype
    ).unsqueeze(-1)
    kys = (by * ns[1]) * torch.fft.fftfreq(
        ns[1], device=cell.device, dtype=cell.dtype
    ).unsqueeze(-1)

    if for_ewald:
        kzs = (bz * ns[2]) * torch.fft.fftfreq(
            ns[2], device=cell.device, dtype=cell.dtype
        ).unsqueeze(-1)
    else:
        kzs = (bz * ns[2]) * torch.fft.rfftfreq(
            ns[2], device=cell.device, dtype=cell.dtype
        ).unsqueeze(-1)

    # then take the cartesian product (all possible combinations, same as meshgrid)
    # via broadcasting (to avoid instantiating intermediates), and sum up
    return kxs[:, None, None] + kys[None, :, None] + kzs[None, None, :]


def generate_kvectors_for_mesh(cell: torch.Tensor, ns: torch.Tensor) -> torch.Tensor:
    """
    Compute all reciprocal space vectors for Fourier space sums.

    This variant is used in combination with **mesh based calculators** using the fast
    fourier transform (FFT) algorithm.

    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param ns: torch.tensor of shape ``(3,)`` and dtype int
        ``ns = [nx, ny, nz]`` contains the number of mesh points in the x-, y- and
        z-direction, respectively. For faster performance during the Fast Fourier
        Transform (FFT) it is recommended to use values of nx, ny and nz that are
        powers of 2.


    :return: torch.tensor of shape ``(nx, ny, nz, 3)`` containing all reciprocal
        space vectors that will be used in the (FFT-based) mesh calculators.
        Note that ``k_vectors[0,0,0] = [0,0,0]`` always is the zero vector.

    .. seealso::

        :func:`generate_kvectors_for_ewald` for a function to be used for Ewald
        calculators.
    """
    return _generate_kvectors(cell=cell, ns=ns, for_ewald=False)


def generate_kvectors_for_ewald(
    cell: torch.Tensor,
    ns: torch.Tensor,
) -> torch.Tensor:
    """
    Compute all reciprocal space vectors for Fourier space sums.

    This variant is used with the **Ewald calculator**, in which the sum over the
    reciprocal space vectors is performed explicitly rather than using the fast Fourier
    transform (FFT) algorithm.

    The main difference with :func:`generate_kvectors_for_mesh` is the shape of the
    output tensor (see documentation on return) and the fact that the full set of
    reciprocal space vectors is returned, rather than the FFT-optimized set that roughly
    contains only half of the vectors.

    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param ns: torch.tensor of shape ``(3,)`` and dtype int
        ``ns = [nx, ny, nz]`` contains the number of mesh points in the x-, y- and
        z-direction, respectively.

    :return: torch.tensor of shape ``(n, 3)`` containing all reciprocal
        space vectors that will be used in the Ewald calculator.
        Note that ``k_vectors[0] = [0,0,0]`` always is the zero vector.

    .. seealso::

        :func:`generate_kvectors_for_mesh` for a function to be used with mesh based
        calculators like PME.
    """
    return _generate_kvectors(cell=cell, ns=ns, for_ewald=True).reshape(-1, 3)



from .kspace_filter import KSpaceFilter, KSpaceKernel, P3MKSpaceFilter
from .kvectors import (
    generate_kvectors_for_ewald,
    generate_kvectors_for_mesh,
    get_ns_mesh,
)
from .math import exp1, gamma, gammaincc_over_powerlaw
from .mesh_interpolator import MeshInterpolator
from .splines import (
    CubicSpline,
    CubicSplineReciprocal,
    compute_second_derivatives,
    compute_spline_ft,
)

__all__ = [
    "CubicSpline",
    "CubicSplineReciprocal",
    "compute_spline_ft",
    "compute_second_derivatives",
    "KSpaceFilter",
    "KSpaceKernel",
    "MeshInterpolator",
    "P3MKSpaceFilter",
    "all_neighbor_indices",
    "distances",
    "generate_kvectors_for_ewald",
    "generate_kvectors_for_mesh",
    "get_ns_mesh",
    "gamma",
    "gammaincc_over_powerlaw",
    "exp1",
]



from typing import Optional

import torch
from torch.special import gammainc

from torchpme.lib import gamma, gammaincc_over_powerlaw

from .potential import Potential


class InversePowerLawPotential(Potential):
    """
    Inverse power-law potentials of the form :math:`1/r^p`.

    Here :math:`r` is a distance parameter and :math:`p` an exponent.

    It can be used to compute:

    1. the full :math:`1/r^p` potential
    2. its short-range (SR) and long-range (LR) parts, the split being determined by a
       length-scale parameter (called "smearing" in the code)
    3. the Fourier transform of the LR part

    :param exponent: the exponent :math:`p` in :math:`1/r^p` potentials
    :param smearing: float or torch.Tensor containing the parameter often called "sigma"
        in publications, which determines the length-scale at which the short-range and
        long-range parts of the naive :math:`1/r^p` potential are separated. For the
        Coulomb potential (:math:`p=1`), this potential can be interpreted as the
        effective potential generated by a Gaussian charge density, in which case this
        smearing parameter corresponds to the "width" of the Gaussian.
    :param: exclusion_radius: float or torch.Tensor containing the length scale
        corresponding to a local environment. See also
        :class:`Potential`.
    :param exclusion_degree: Controls the sharpness of the transition in the cutoff function
        applied within the ``exclusion_radius``. The cutoff is computed as a raised cosine
        with exponent ``exclusion_degree``
    """

    def __init__(
        self,
        exponent: int,
        smearing: Optional[float] = None,
        exclusion_radius: Optional[float] = None,
        exclusion_degree: int = 1,
    ):
        super().__init__(smearing, exclusion_radius, exclusion_degree)

        # function call to check the validity of the exponent
        gammaincc_over_powerlaw(exponent, torch.tensor(1.0))
        self.register_buffer("exponent", torch.tensor(exponent, dtype=torch.float64))

    @torch.jit.export
    def from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        """
        Full :math:`1/r^p` potential as a function of :math:`r`.

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        return torch.pow(dist, -self.exponent)

    @torch.jit.export
    def lr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        """
        Long range of the range-separated :math:`1/r^p` potential.

        Used to subtract out the interior contributions after computing the LR part in
        reciprocal (Fourier) space.

        For the Coulomb potential, this would return (note that the only change between
        the SR and LR parts is the fact that erfc changes to erf)

        .. code-block:: python

            potential = erf(dist / sqrt(2) / smearing) / dist

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range contribution without specifying `smearing`."
            )

        x = 0.5 * dist**2 / self.smearing**2
        peff = self.exponent / 2
        prefac = 1.0 / (2 * self.smearing**2) ** peff
        return prefac * gammainc(peff, x) / x**peff

    @torch.jit.export
    def lr_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        r"""
        Fourier transform of the LR part potential in terms of :math:`\mathbf{k^2}`.

        :param k_sq: torch.tensor containing the squared lengths (2-norms) of the wave
            vectors k at which the Fourier-transformed potential is to be evaluated
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range kernel without specifying `smearing`."
            )

        peff = (3 - self.exponent) / 2
        prefac = (
            torch.pi**1.5 / gamma(self.exponent / 2) * (2 * self.smearing**2) ** peff
        )
        x = 0.5 * self.smearing**2 * k_sq

        # The k=0 term often needs to be set separately since for exponents p<=3
        # dimension, there is a divergence to +infinity. Setting this value manually
        # to zero physically corresponds to the addition of a uniform background charge
        # to make the system charge-neutral. For p>3, on the other hand, the
        # Fourier-transformed LR potential does not diverge as k->0, and one
        # could instead assign the correct limit. This is not implemented for now
        # for consistency reasons.
        masked = torch.where(x == 0, 1.0, x)  # avoid NaNs in backwards, see Coulomb
        return torch.where(
            k_sq == 0, 0.0, prefac * gammaincc_over_powerlaw(self.exponent, masked)
        )

    def self_contribution(self) -> torch.Tensor:
        # self-correction for 1/r^p potential
        if self.smearing is None:
            raise ValueError(
                "Cannot compute self contribution without specifying `smearing`."
            )
        phalf = self.exponent / 2
        return 1 / gamma(phalf + 1) / (2 * self.smearing**2) ** phalf

    def background_correction(self) -> torch.Tensor:
        # "charge neutrality" correction for 1/r^p potential diverges for exponent p = 3
        # and is not needed for p > 3 , so we set it to zero (see in
        # https://doi.org/10.48550/arXiv.2412.03281 SI section)
        if self.smearing is None:
            raise ValueError(
                "Cannot compute background correction without specifying `smearing`."
            )
        if self.exponent >= 3:
            return torch.zeros_like(self.smearing)
        prefac = torch.pi**1.5 * (2 * self.smearing**2) ** ((3 - self.exponent) / 2)
        prefac /= (3 - self.exponent) * gamma(self.exponent / 2)
        return prefac

    self_contribution.__doc__ = Potential.self_contribution.__doc__
    background_correction.__doc__ = Potential.background_correction.__doc__



import torch
from torch.special import gammaln


def gamma(x: torch.Tensor) -> torch.Tensor:
    """
    (Complete) Gamma function.

    pytorch has not implemented the commonly used (complete) Gamma function. We define
    it in a custom way to make autograd work as in
    https://discuss.pytorch.org/t/is-there-a-gamma-function-in-pytorch/17122
    """
    return torch.exp(gammaln(x))


class _CustomExp1(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        # this implementation is inspired by the one in scipy:
        # https://github.com/scipy/scipy/blob/34d91ce06d4d05e564b79bf65288284247b1f3e3/scipy/special/xsf/expint.h#L22
        ctx.save_for_backward(x)

        # Constants
        SCIPY_EULER = (
            0.577215664901532860606512090082402431  # Euler-Mascheroni constant
        )
        inf = torch.inf

        # Handle case when x == 0
        result = torch.full_like(x, inf)
        mask = x > 0

        # Compute for x <= 1
        x_small = x[mask & (x <= 1)]
        if x_small.numel() > 0:
            e1 = torch.ones_like(x_small)
            r = torch.ones_like(x_small)
            for k in range(1, 26):
                r = -r * k * x_small / (k + 1.0) ** 2
                e1 += r
                if torch.all(torch.abs(r) <= torch.abs(e1) * 1e-15):
                    break
            result[mask & (x <= 1)] = -SCIPY_EULER - torch.log(x_small) + x_small * e1

        # Compute for x > 1
        x_large = x[mask & (x > 1)]
        if x_large.numel() > 0:
            m = 20 + (80.0 / x_large).to(torch.int32)
            t0 = torch.zeros_like(x_large)
            for k in range(m.max(), 0, -1):
                t0 = k / (1.0 + k / (x_large + t0))
            t = 1.0 / (x_large + t0)
            result[mask & (x > 1)] = torch.exp(-x_large) * t

        return result

    @staticmethod
    def backward(ctx, grad_output):
        (x,) = ctx.saved_tensors
        return -grad_output * torch.exp(-x) / x


def exp1(x):
    r"""
    Exponential integral E1.

    For a real number :math:`x > 0` the exponential integral can be defined as

    .. math::

        E1(x) = \int_{x}^{\infty} \frac{e^{-t}}{t} dt

    :param x: Input tensor (x > 0)
    :return: Exponential integral E1(x)
    """
    return _CustomExp1.apply(x)


def gammaincc_over_powerlaw(exponent: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
    """
    Compute the regularized incomplete gamma function complement for integer exponents.

    :param exponent: Exponent of the power law
    :param z: Value at which to evaluate the function
    :return: Regularized incomplete gamma function complement
    """
    if exponent == 1:
        return torch.exp(-z) / z
    if exponent == 2:
        return torch.sqrt(torch.pi / z) * torch.erfc(torch.sqrt(z))
    if exponent == 3:
        return exp1(z)
    if exponent == 4:
        return 2 * (
            torch.exp(-z) - torch.sqrt(torch.pi * z) * torch.erfc(torch.sqrt(z))
        )
    if exponent == 5:
        return torch.exp(-z) - z * exp1(z)
    if exponent == 6:
        return (
            (2 - 4 * z) * torch.exp(-z)
            + 4 * torch.sqrt(torch.pi * z**3) * torch.erfc(torch.sqrt(z))
        ) / 3
    raise ValueError(f"Unsupported exponent: {exponent}")



import math
from typing import Any
from warnings import warn

import torch

from ..calculators import EwaldCalculator
from .tuner import GridSearchTuner, TuningErrorBounds


def tune_ewald(
    charges: torch.Tensor,
    cell: torch.Tensor,
    positions: torch.Tensor,
    cutoff: float,
    neighbor_indices: torch.Tensor,
    neighbor_distances: torch.Tensor,
    full_neighbor_list: bool = False,
    prefactor: float = 1.0,
    exponent: int = 1,
    ns_lo: int = 1,
    ns_hi: int = 14,
    accuracy: float = 1e-3,
) -> tuple[float, dict[str, Any], float]:
    r"""
    Find the optimal parameters for :class:`torchpme.EwaldCalculator`.

    .. note::

        The :func:`torchpme.tuning.ewald.EwaldErrorBounds.forward` method takes floats
        as the input, in order to be in consistency with the rest of the package --
        these parameters are always ``float`` but not ``torch.Tensor``. This design,
        however, prevents the utilization of ``torch.autograd`` and other ``torch``
        features. To take advantage of these features, one can use the
        :func:`torchpme.tuning.ewald.EwaldErrorBounds.err_rspace` and
        :func:`torchpme.tuning.ewald.EwaldErrorBounds.err_kspace`, which takes
        ``torch.Tensor`` as parameters.

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.
    :param cutoff: float, cutoff distance for the neighborlist
    :param exponent: :math:`p` in :math:`1/r^p` potentials, currently only :math:`p=1`
        is supported
    :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
        which the potential should be computed in real space.
    :param neighbor_distances: torch.tensor with the pair distances of the neighbors for
        which the potential should be computed in real space.
    :param full_neighbor_list: If set to :py:obj:`True`, a "full" neighbor list
        is expected as input. This means that each atom pair appears twice. If
        set to :py:obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    :param ns_lo: Minimum number of spatial resolution along each axis
    :param ns_hi: Maximum number of spatial resolution along each axis
    :param accuracy: Recomended values for a balance between the accuracy and speed is
        :math:`10^{-3}`. For more accurate results, use :math:`10^{-6}`.

    :return: Tuple containing a float of the optimal smearing for the :class:
        `CoulombPotential`, and a dictionary with the parameters for
        :class:`EwaldCalculator`, and the timing of this set of parameters.

    Example
    -------
    >>> import torch
    >>> positions = torch.tensor([[0.0, 0.0, 0.0], [0.4, 0.4, 0.4]])
    >>> charges = torch.tensor([[1.0], [-1.0]])
    >>> cell = torch.eye(3)
    >>> neighbor_distances = torch.tensor(
    ...     [0.9381, 0.9381, 0.8246, 0.9381, 0.8246, 0.8246, 0.6928],
    ... )
    >>> neighbor_indices = torch.tensor(
    ...     [[0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1]]
    ... )
    >>> smearing, parameter, timing = tune_ewald(
    ...     charges,
    ...     cell,
    ...     positions,
    ...     cutoff=1.0,
    ...     neighbor_distances=neighbor_distances,
    ...     neighbor_indices=neighbor_indices,
    ...     accuracy=1e-1,
    ... )

    """
    # if cell is 0 `min_dimension` will be zero as well; we raise a propper error later
    min_dimension = float(torch.min(torch.linalg.norm(cell, dim=1)))
    params = [{"lr_wavelength": min_dimension / ns} for ns in range(ns_lo, ns_hi + 1)]

    tuner = GridSearchTuner(
        charges=charges,
        cell=cell,
        positions=positions,
        cutoff=cutoff,
        exponent=exponent,
        neighbor_indices=neighbor_indices,
        neighbor_distances=neighbor_distances,
        full_neighbor_list=full_neighbor_list,
        prefactor=prefactor,
        calculator=EwaldCalculator,
        error_bounds=EwaldErrorBounds(charges=charges, cell=cell, positions=positions),
        params=params,
    )
    smearing = tuner.estimate_smearing(accuracy)
    errs, timings = tuner.tune(accuracy)

    # There are multiple errors below the accuracy, return the one with the shortest
    # calculation time. The timing of those parameters leading to an higher error than
    # the accuracy are set to infinity
    if any(err < accuracy for err in errs):
        return smearing, params[timings.index(min(timings))], min(timings)
    # No parameter meets the requirement, return the one with the smallest error
    warn(
        f"No parameter meets the accuracy requirement.\n"
        f"Returning the parameter with the smallest error, which is {min(errs)}.\n",
        stacklevel=1,
    )
    return smearing, params[errs.index(min(errs))], timings[errs.index(min(errs))]


class EwaldErrorBounds(TuningErrorBounds):
    r"""
    Error bounds for :class:`torchpme.calculators.ewald.EwaldCalculator`.

        .. math::
            \text{Error}_{\text{total}} = \sqrt{\text{Error}_{\text{real space}}^2 +
            \text{Error}_{\text{Fourier space}}^2

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.

    Example
    -------
    >>> import torch
    >>> positions = torch.tensor([[0.0, 0.0, 0.0], [0.4, 0.4, 0.4]])
    >>> charges = torch.tensor([[1.0], [-1.0]])
    >>> cell = torch.eye(3)
    >>> error_bounds = EwaldErrorBounds(charges, cell, positions)
    >>> print(error_bounds(smearing=1.0, lr_wavelength=0.5, cutoff=4.4))
    tensor(8.4304e-05)

    """

    def __init__(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
    ):
        super().__init__(charges, cell, positions)

        self.volume = torch.abs(torch.det(cell))
        self.sum_squared_charges = (charges**2).sum()
        self.prefac = 2 * self.sum_squared_charges / math.sqrt(len(positions))
        self.cell = cell
        self.positions = positions

    def err_kspace(
        self, smearing: torch.Tensor, lr_wavelength: torch.Tensor
    ) -> torch.Tensor:
        """
        The Fourier space error of Ewald.

        :param smearing: see :class:`torchpme.EwaldCalculator` for details
        :param lr_wavelength: see :class:`torchpme.EwaldCalculator` for details
        """
        return (
            self.prefac**0.5
            / smearing
            / torch.sqrt((2 * torch.pi) ** 2 * self.volume / (lr_wavelength) ** 0.5)
            * torch.exp(-((2 * torch.pi) ** 2) * smearing**2 / (lr_wavelength))
        )

    def err_rspace(self, smearing: torch.Tensor, cutoff: torch.Tensor) -> torch.Tensor:
        """
        The real space error of Ewald.

        :param smearing: see :class:`torchpme.EwaldCalculator` for details
        :param lr_wavelength: see :class:`torchpme.EwaldCalculator` for details
        """
        return (
            self.prefac
            / torch.sqrt(cutoff * self.volume)
            * torch.exp(-(cutoff**2) / 2 / smearing**2)
        )

    def forward(
        self, smearing: float, lr_wavelength: float, cutoff: float
    ) -> torch.Tensor:
        r"""
        Calculate the error bound of Ewald.

        :param smearing: see :class:`torchpme.EwaldCalculator` for details
        :param lr_wavelength: see :class:`torchpme.EwaldCalculator` for details
        :param cutoff: see :class:`torchpme.EwaldCalculator` for details
        """
        smearing = torch.tensor(smearing)
        lr_wavelength = torch.tensor(lr_wavelength)
        cutoff = torch.tensor(cutoff)
        return torch.sqrt(
            self.err_kspace(smearing, lr_wavelength) ** 2
            + self.err_rspace(smearing, cutoff) ** 2
        )



from .calculator import Calculator
from .calculator_dipole import CalculatorDipole
from .ewald import EwaldCalculator
from .p3m import P3MCalculator
from .pme import PMECalculator

__all__ = [
    "Calculator",
    "EwaldCalculator",
    "P3MCalculator",
    "PMECalculator",
    "CalculatorDipole",
]



import math
import time
from typing import Optional

import torch

from .._utils import _validate_parameters
from ..calculators import Calculator
from ..potentials import InversePowerLawPotential


class TuningErrorBounds(torch.nn.Module):
    """
    Base class for error bounds. This class calculates the real space error and the
    Fourier space error based on the error formula. This class is used in the tuning
    process. It can also be used with the :class:`torchpme.tuning.tuner.TunerBase` to
    build up a custom parameter tuner.

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.
    """

    def __init__(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
    ):
        super().__init__()
        self._charges = charges
        self._cell = cell
        self._positions = positions

    def forward(self, *args, **kwargs):
        return self.error(*args, **kwargs)

    def error(self, *args, **kwargs):
        raise NotImplementedError


class TunerBase:
    """
    Base class defining the interface for a parameter tuner.

    This class provides a framework for tuning the parameters of a calculator. The class
    itself supports estimating the ``smearing`` from the real space cutoff based on the
    real space error formula. The :func:`TunerBase.tune` defines the interface for a
    sophisticated tuning process, which takes a value of the desired accuracy.

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.
    :param cutoff: real space cutoff, serves as a hyperparameter here.
    :param calculator: the calculator to be tuned
    :param exponent: exponent of the potential, only exponent = 1 is supported
    :param full_neighbor_list: If set to :py:obj:`True`, a "full" neighbor list
        is expected as input. This means that each atom pair appears twice. If
        set to :py:obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.

    Example
    -------
    >>> import torch
    >>> import torchpme
    >>> positions = torch.tensor([[0.0, 0.0, 0.0], [0.4, 0.4, 0.4]])
    >>> charges = torch.tensor([[1.0], [-1.0]])
    >>> cell = torch.eye(3)
    >>> tuner = TunerBase(charges, cell, positions, 4.4, torchpme.EwaldCalculator)
    >>> smearing = tuner.estimate_smearing(1e-3)
    >>> print(smearing)
    1.1069526756106463

    """

    def __init__(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
        cutoff: float,
        calculator: type[Calculator],
        exponent: int = 1,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
    ):
        if exponent != 1:
            raise NotImplementedError(
                f"Only exponent = 1 is supported but got {exponent}."
            )

        _validate_parameters(
            charges=charges,
            cell=cell,
            positions=positions,
            neighbor_indices=torch.tensor([[0, 1]], device=positions.device),
            neighbor_distances=torch.tensor(
                [1.0], device=positions.device, dtype=positions.dtype
            ),
            smearing=1.0,  # dummy value because; always have range-seperated potentials
        )
        self.charges = charges
        self.cell = cell
        self.positions = positions
        self.cutoff = cutoff
        self.calculator = calculator
        self.exponent = exponent
        self.full_neighbor_list = full_neighbor_list
        self.prefactor = prefactor

        self._smearing_esti_prefac = (
            2 * float((charges**2).sum()) / math.sqrt(len(positions))
        )

    def tune(self, accuracy: float = 1e-3):
        raise NotImplementedError

    def estimate_smearing(
        self,
        accuracy: float,
    ) -> float:
        """
        Estimate the smearing based on the error formula of the real space. The
        smearing is set as leading to a real space error of ``accuracy/4``.

        :param accuracy: a float, the desired accuracy
        :return: a float, the estimated smearing
        """
        if not isinstance(accuracy, float):
            raise ValueError(f"'{accuracy}' is not a float.")
        ratio = math.sqrt(
            -2
            * math.log(
                accuracy
                / 2
                / self._smearing_esti_prefac
                * math.sqrt(self.cutoff * float(torch.abs(self.cell.det())))
            )
        )
        smearing = self.cutoff / ratio

        return float(smearing)

    @staticmethod
    def filter_neighbors(
        cutoff: float, neighbor_indices: torch.Tensor, neighbor_distances: torch.Tensor
    ):
        """
        Filter neighbor indices and distances based on a user given cutoff. This allows
        users pre-computing the neighbor list with a larger cutoff and then filtering
        the neighbors based on a smaller cutoff, leading to a faster tuning on the
        cutoff.

        :param cutoff: real space cutoff
        :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
            which the potential should be computed in real space.
        :param neighbor_distances: torch.tensor with the pair distances of the neighbors
            for which the potential should be computed in real space.
        """
        filter_idx = torch.where(neighbor_distances < cutoff)
        return neighbor_indices[filter_idx], neighbor_distances[filter_idx]


class GridSearchTuner(TunerBase):
    """
    Tuner using grid search.

    The tuner uses the error formula to estimate the error of a given parameter set. If
    the error is smaller than the accuracy, the timing is measured and returned. If the
    error is larger than the accuracy, the timing is set to infinity and the parameter
    is skipped.

    .. note::

        The cutoff is treated as a hyperparameter here. In case one wants to tune the
        cutoff, one could instantiate the tuner with different cutoff values and
        manually pick the best from the tuning results.

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.
    :param cutoff: real space cutoff, serves as a hyperparameter here.
    :param calculator: the calculator to be tuned
    :param error_bounds: error bounds for the calculator
    :param params: list of Fourier space parameter sets for which the error is estimated
    :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
        which the potential should be computed in real space.
    :param neighbor_distances: torch.tensor with the pair distances of the neighbors for
        which the potential should be computed in real space.
    :param full_neighbor_list: If set to :py:obj:`True`, a "full" neighbor list
        is expected as input. This means that each atom pair appears twice. If
        set to :py:obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    :param exponent: exponent of the potential, only exponent = 1 is supported
    """

    def __init__(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
        cutoff: float,
        calculator: type[Calculator],
        error_bounds: type[TuningErrorBounds],
        params: list[dict],
        neighbor_indices: torch.Tensor,
        neighbor_distances: torch.Tensor,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
        exponent: int = 1,
    ):
        super().__init__(
            charges=charges,
            cell=cell,
            positions=positions,
            cutoff=cutoff,
            calculator=calculator,
            exponent=exponent,
            full_neighbor_list=full_neighbor_list,
            prefactor=prefactor,
        )
        self.error_bounds = error_bounds
        self.params = params
        neighbor_indices, neighbor_distances = self.filter_neighbors(
            cutoff, neighbor_indices, neighbor_distances
        )
        self.time_func = TuningTimings(
            charges,
            cell,
            positions,
            neighbor_indices,
            neighbor_distances,
            True,
        )

    def tune(self, accuracy: float = 1e-3) -> tuple[list[float], list[float]]:
        """
        Estimate the error and timing for each parameter set. Only parameters for
        which the error is smaller than the accuracy are timed, the others' timing is
        set to infinity.

        :param accuracy: a float, the desired accuracy
        :return: a list of errors and a list of timings
        """
        if not isinstance(accuracy, float):
            raise ValueError(f"'{accuracy}' is not a float.")
        smearing = self.estimate_smearing(accuracy)
        param_errors = []
        param_timings = []
        for param in self.params:
            error = self.error_bounds(smearing=smearing, cutoff=self.cutoff, **param)  # type: ignore[call-arg]
            param_errors.append(float(error))
            # only computes timings for parameters that meet the accuracy requirements
            param_timings.append(
                self._timing(smearing, param) if error <= accuracy else float("inf")
            )

        return param_errors, param_timings

    def _timing(self, smearing: float, k_space_params: dict):
        calculator = self.calculator(
            potential=InversePowerLawPotential(
                exponent=self.exponent,  # but only exponent = 1 is supported
                smearing=smearing,
            ),
            full_neighbor_list=self.full_neighbor_list,
            prefactor=self.prefactor,
            **k_space_params,
        )
        calculator.to(device=self.positions.device, dtype=self.positions.dtype)
        return self.time_func(calculator)


class TuningTimings(torch.nn.Module):
    """
    Class for timing a calculator.

    The class estimates the average execution time of a given calculater after several
    warmup runs. The class takes the information of the structure that one wants to
    benchmark on, and the configuration of the timing process as inputs.

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.
    :param cutoff: real space cutoff, serves as a hyperparameter here.
    :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
        which the potential should be computed in real space.
    :param neighbor_distances: torch.tensor with the pair distances of the neighbors for
        which the potential should be computed in real space.
    :param n_repeat: number of times to repeat to estimate the average timing
    :param n_warmup: number of warmup runs, recommended to be at least 4
    :param run_backward: whether to run the backward pass
    """

    def __init__(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
        neighbor_indices: torch.Tensor,
        neighbor_distances: torch.Tensor,
        n_repeat: int = 4,
        n_warmup: int = 4,
        run_backward: Optional[bool] = True,
    ):
        super().__init__()

        _validate_parameters(
            charges=charges,
            cell=cell,
            positions=positions,
            neighbor_indices=neighbor_indices,
            neighbor_distances=neighbor_distances,
            smearing=1.0,  # dummy value because; always have range-seperated potentials
        )

        self.charges = charges
        self.cell = cell
        self.positions = positions
        self.n_repeat = n_repeat
        self.n_warmup = n_warmup
        self.run_backward = run_backward
        self.neighbor_indices = neighbor_indices
        self.neighbor_distances = neighbor_distances

    def forward(self, calculator: torch.nn.Module):
        """
        Estimate the execution time of a given calculator for the structure
        to be used as benchmark.

        :param calculator: the calculator to be tuned
        :return: a float, the average execution time
        """
        # measure time
        execution_time = 0.0

        for _ in range(self.n_repeat + self.n_warmup):
            if _ == self.n_warmup:
                execution_time = 0.0
            positions = self.positions.clone()
            cell = self.cell.clone()
            charges = self.charges.clone()
            # nb - this won't compute gradiens involving the distances
            if self.run_backward:
                positions.requires_grad_(True)
                cell.requires_grad_(True)
                charges.requires_grad_(True)
            execution_time -= time.monotonic()
            result = calculator.forward(
                positions=positions,
                charges=charges,
                cell=cell,
                neighbor_indices=self.neighbor_indices,
                neighbor_distances=self.neighbor_distances,
            )
            value = result.sum()
            if self.run_backward:
                value.backward(retain_graph=True)

            execution_time += time.monotonic()

        return execution_time / self.n_repeat



from typing import Optional

import torch


class CubicSpline(torch.nn.Module):
    r"""
    Cubic spline calculator.

    Class implementing a cubic spline for a real-valued function.

    :param x_points: Abscissas of the splining points for the real-space function
    :param y_points: Ordinates of the splining points for the real-space function
    """

    def __init__(self, x_points: torch.Tensor, y_points: torch.Tensor):
        super().__init__()

        # stores grid information needed to compute the
        self.x_points = x_points
        self.y_points = y_points
        self.d2y_points = compute_second_derivatives(x_points, y_points)
        self._intervals = self.x_points[1:] - self.x_points[:-1]
        self._h2over6 = self._intervals**2 / 6

    def forward(self, x: torch.Tensor):
        """
        Evaluates the spline at the points provided.

        :param x: One or more positions to evaluate the splined function.
        """
        # Calculate the spline for each x
        i = torch.searchsorted(self.x_points, x, right=True) - 1
        i = torch.clamp(i, 0, len(self.x_points) - 2)

        h = self._intervals[i]
        a = (self.x_points[i + 1] - x) / h
        b = (x - self.x_points[i]) / h
        h2over6 = self._h2over6[i]
        return a * (
            self.y_points[i] + (a * a - 1) * self.d2y_points[i] * h2over6
        ) + b * (self.y_points[i + 1] + (b * b - 1) * self.d2y_points[i + 1] * h2over6)


class CubicSplineReciprocal(torch.nn.Module):
    r"""
    Reciprocal-axis cubic spline calculator.

    Computes a spline on a :math:`1/x` grid, "extending" it so that
    it converges smoothly to zero as :math:`x\rightarrow\infty`.
    The function parameters are still provided as (x,y) points, but the
    interpolation is performed internally using :math:`1/x`, so the radial grid
    should only contain strictly-positive values.

    :param x_points: Abscissas of the splining points for the real-space function.
        Must be strictly larger than zero. It is recommended for the smaller value
        to be much smaller than the minimum expected distance between atoms.
    :param y_points: Ordinates of the splining points for the real-space function
    :param y_at_zero: Value to be returned when called for an argument of zero.
        Also uses a direct interpolation to "fill in the blanks". Defaults to the
        value of ``y_points[0]``.
    """

    def __init__(
        self,
        x_points: torch.Tensor,
        y_points: torch.Tensor,
        y_at_zero: Optional[torch.Tensor] = None,
    ):
        super().__init__()

        # compute on a inverse grid
        ix_points = torch.cat(
            [
                torch.zeros((1,), dtype=x_points.dtype, device=x_points.device),
                torch.reciprocal(torch.flip(x_points, dims=[0])),
            ],
            dim=0,
        )
        iy_points = torch.cat(
            [
                torch.zeros((1,), dtype=x_points.dtype, device=x_points.device),
                torch.flip(y_points, dims=[0]),
            ],
            dim=0,
        )
        self._rev_spline = CubicSpline(ix_points, iy_points)

        # defaults to the lowest value in the input
        if y_at_zero is None:
            y_at_zero = y_points[0]
        self._y_at_zero = y_at_zero
        # direct mini-spline to fill the gap between the lowest grid point and zero
        self._zero_spline = CubicSpline(
            torch.tensor(
                [0.0, x_points[0], x_points[1]],
                dtype=x_points.dtype,
                device=x_points.device,
            ),
            torch.tensor(
                [self._y_at_zero, y_points[0], y_points[1]],
                dtype=x_points.dtype,
                device=x_points.device,
            ),
        )

    def forward(self, x: torch.Tensor):
        """Computes by reversing the inputs, checking for safety."""
        safe_x = torch.where(
            x < self._zero_spline.x_points[1], self._zero_spline.x_points[1], x
        )
        return torch.where(
            x < self._zero_spline.x_points[1],
            self._zero_spline(x),
            self._rev_spline(torch.reciprocal(safe_x)),
        )


def _solve_tridiagonal(a, b, c, d):
    """
    Helper function to solve a tri-diagonal linear problem.

    a = torch.zeros(n)  # Sub-diagonal (a[1..n-1])
    b = torch.zeros(n)  # Main diagonal (b[0..n-1])
    c = torch.zeros(n)  # Super-diagonal (c[0..n-2])
    d = torch.zeros(n)  # Right-hand side (d[0..n-1])
    """
    n = len(d)
    # Create copies to avoid modifying the original arrays
    c_prime = torch.zeros_like(d)
    d_prime = torch.zeros_like(d)

    # Initial coefficients
    c_prime[0] = c[0] / b[0]
    d_prime[0] = d[0] / b[0]

    # Forward sweep
    for i in range(1, n):
        denom = b[i] - a[i] * c_prime[i - 1]
        c_prime[i] = c[i] / denom if i < n - 1 else 0
        d_prime[i] = (d[i] - a[i] * d_prime[i - 1]) / denom

    # Backward substitution
    x = torch.zeros_like(d)
    x[-1] = d_prime[-1]
    for i in reversed(range(n - 1)):
        x[i] = d_prime[i] - c_prime[i] * x[i + 1]
    return x


def compute_second_derivatives(
    x_points: torch.Tensor,
    y_points: torch.Tensor,
):
    """
    Computes second derivatives given the grid points of a cubic spline.

    :param x_points: Abscissas of the splining points for the real-space function
    :param y_points: Ordinates of the splining points for the real-space function

    :return: The second derivatives for the spline points
    """
    # Do the calculation in float64 if required
    x = x_points
    y = y_points

    # Calculate intervals
    intervals = x[1:] - x[:-1]
    dy = (y[1:] - y[:-1]) / intervals

    n = len(x)
    a = torch.zeros_like(x)  # Sub-diagonal (a[1..n-1])
    b = torch.zeros_like(x)  # Main diagonal (b[0..n-1])
    c = torch.zeros_like(x)  # Super-diagonal (c[0..n-2])
    d = torch.zeros_like(x)  # Right-hand side (d[0..n-1])

    # Natural spline boundary conditions
    b[0] = 1
    d[0] = 0  # Second derivative at the first point is zero
    b[-1] = 1
    d[-1] = 0  # Second derivative at the last point is zero

    # Fill the diagonals and right-hand side
    for i in range(1, n - 1):
        a[i] = intervals[i - 1] / 6
        b[i] = (intervals[i - 1] + intervals[i]) / 3
        c[i] = intervals[i] / 6
        d[i] = dy[i] - dy[i - 1]

    return _solve_tridiagonal(a, b, c, d)

    # Converts back to the original dtype


def compute_spline_ft(
    k_points: torch.Tensor,
    x_points: torch.Tensor,
    y_points: torch.Tensor,
    d2y_points: torch.Tensor,
):
    r"""
    Computes the Fourier transform of a splined radial function.

    Evaluates the integral

    .. math::

        \hat{f}(k) =4\pi\int \mathrm{d}r \frac{\sin k r}{k} r f(r)

    where :math:`f(r)` is expressed as a cubic spline. The function
    also includes a tail correction to continue the integral beyond
    the last splined point, assuming that the function converges to
    zero at infinity.

    :param k_points:  Points on which the Fourier kernel should be
        computed. It is a good idea to take them to be
        :math:`2\pi/x` based on the real-space ``x_points``
    :param x_points: Abscissas of the splining points for the real-space function
    :param y_points: Ordinates of the splining points for the real-space function
    :param d2y_points:  Second derivatives for the spline points

    :return: The radial Fourier transform :math:`\hat{f}(k)` computed
        at the ``k_points`` provided.
    """
    # the expression contains the cosine integral special function, that
    # is only available in scipy
    try:
        import scipy.special
    except ImportError as err:
        raise ImportError(
            "Computing the Fourier-domain kernel based on a spline requires scipy"
        ) from err

    # chooses precision for the FT evaluation
    dtype = x_points.dtype

    # broadcast to compute at once on all k values.
    # all these are terms that enter the analytical integral.
    # might be possible to write this in a more concise way, but
    # this works and is reasonably numerically stable, so it will do
    k = k_points.reshape(-1, 1).to(dtype)  # target k's
    ri = x_points[torch.newaxis, :-1].to(dtype)  # radial grid points
    yi = y_points[torch.newaxis, :-1].to(dtype)  # radial grid values
    d2yi = d2y_points[torch.newaxis, :-1].to(dtype)  # radial spline second derivatives
    # corresponding increments
    dr = (x_points[torch.newaxis, 1:] - x_points[torch.newaxis, :-1]).to(dtype)
    dy = (y_points[torch.newaxis, 1:] - y_points[torch.newaxis, :-1]).to(dtype)
    dd2y = (d2y_points[torch.newaxis, 1:] - d2y_points[torch.newaxis, :-1]).to(dtype)
    # trig functions at grid points
    coskx = torch.cos(k * ri)
    sinkx = torch.sin(k * ri)
    # trig function increments, computed with trig identities for stability
    # cos r+dr - cos r
    dcoskx = 2 * torch.sin(k * dr / 2) * torch.sin(k * (dr / 2 + ri))
    # sin r+dr - cos r
    dsinkx = -2 * torch.sin(k * dr / 2) * torch.cos(k * (dr / 2 + ri))

    # this monstruous expression computes, for each interval in the spline,
    # \int_{r_i}^{r_{i+1}} spline_i(r) f(k,r) dr using the coefficients of
    # the spline (f(k,r) is the radial FT coefficient, see the expresison in
    # the docstring).
    # the resulting integral (use a CAS to compute it!) can be written in terms
    # of the spline point at i, the increments, and the trigonometric functions.
    # it formally contains a 1/k^6 pole, that is however removable, because the
    # numerator also goes to zero as k^6. this is addressed in several different
    # ways. (1) the expression is made more stable for small k by casting it in a
    # Horner form; (2) the first term contains the difference
    # of two cosines (at i and i+1), but is computed with a trigonometric identity
    # (see the definition of dcoskx) to avoid the 1-k^2 form of the bare cosines
    # (3) the k->0 limit is computed analytically and used if one point is strictly
    # zero  (4) division by k^6 is delayed, and done conditionally.
    ft_interval = 24 * dcoskx * dd2y + k * (
        6 * dsinkx * (3 * d2yi * dr + dd2y * (4 * dr + ri))
        - 24 * dd2y * dr * sinkx
        + k
        * (
            6 * coskx * dr * (3 * d2yi * dr + dd2y * (2 * dr + ri))
            - 2
            * dcoskx
            * (6 * dy + dr * ((6 * d2yi + 5 * dd2y) * dr + 3 * (d2yi + dd2y) * ri))
            + k
            * (
                dr
                * (
                    12 * dy
                    + 3 * d2yi * dr * (dr + 2 * ri)
                    + dd2y * dr * (2 * dr + 3 * ri)
                )
                * sinkx
                + dsinkx
                * (
                    -6 * dy * ri
                    - 3 * d2yi * dr**2 * (dr + ri)
                    - 2 * dd2y * dr**2 * (dr + ri)
                    - 6 * dr * (2 * dy + yi)
                )
                + k
                * (
                    6 * dcoskx * dr * (dr + ri) * (dy + yi)
                    + coskx * (6 * dr * ri * yi - 6 * dr * (dr + ri) * (dy + yi))
                )
            )
        )
    )

    # especially for Coulomb-like integrals, no matter how far we push the splining
    # in real space, the tail matters, so we compute it separately. to do this
    # stably and acurately, we build the tail as a spline in 1/r (using the last two)
    # points of the spline) and use an analytical expression for the resulting
    # integral from the last point to infinity
    tail_d2y = compute_second_derivatives(
        torch.tensor([0, 1 / x_points[-1], 1 / x_points[-2]]),
        torch.tensor([0, y_points[-1], y_points[-2]]),
    )

    r0 = x_points[-1]
    y0 = y_points[-1]
    d2y0 = tail_d2y[1]

    # to numpy and back again. this function is only called at initialization
    # time, and so we can live with losing some time here
    cosint = torch.from_numpy(
        scipy.special.sici((k * r0).detach().cpu().numpy())[1]
    ).to(dtype=dr.dtype, device=dr.device)

    # this is the tail contribution multiplied by k**2 to remove the singularity
    tail = (
        -2
        * torch.pi
        * (
            (d2y0 - 6 * r0**2 * y0) * torch.cos(k * r0)
            + d2y0 * k * r0 * (k * r0 * cosint - torch.sin(k * r0))
        )
    ) / (3.0 * r0)

    ft_sum = torch.pi * 2 / 3 * torch.sum(ft_interval / dr, axis=1).reshape(-1, 1)
    # for the interval integrals, there is a finite k-> 0 limit (i.e. the k^-6 divergence cancels)
    ft_limit = torch.sum(
        -(
            dr
            * torch.pi
            * (
                3 * d2yi * dr**2 * (3 * dr**2 + 10 * dr * ri + 10 * ri**2)
                + dd2y * dr**2 * (5 * dr**2 + 16 * dr * ri + 15 * ri**2)
                - 30
                * (
                    6 * ri**2 * (dy + 2 * yi)
                    + 4 * dr * ri * (2 * dy + 3 * yi)
                    + dr**2 * (3 * dy + 4 * yi)
                )
            )
        )
        / 90,
        axis=1,
    )

    safe_k = torch.where(k == 0, 1.0, k)
    return (
        torch.where(
            k == 0,
            ft_limit,
            ft_sum / safe_k**6 + tail / safe_k**2,
        )
        .reshape(k_points.shape)
        .to(k_points.dtype)
    )



from typing import Optional

import torch

from .potential import Potential


class PotentialDipole(torch.nn.Module):
    r"""
    Pair potential energy function between point dipoles.

    The intercation is described as

    .. math::

        V(\vec{r}) = \frac{(\vec{\mu}_i \cdot \vec{\mu}_j)}{r^3} -
            \frac{3  (\vec{\mu}_i \cdot \vec{r})  (\vec{\mu}_j \cdot \vec{r}) }{r^5}

    where :math:`r=|\vec{r}|`.

    :param smearing: float or torch.Tensor containing the parameter often called "sigma"
        in publications, which determines the length-scale at which the short-range and
        long-range parts of the naive :math:`1/r` potential are separated. The smearing
        parameter corresponds to the "width" of a Gaussian smearing of the particle
        density.
    :param exclusion_radius: A length scale that defines a *local environment* within
        which the potential should be smoothly zeroed out, as it will be described by a
        separate model.
    :param epsilon: Dielectric constant of the medium in which the dipoles are embedded.
    :param exclusion_degree: Controls the sharpness of the transition in the cutoff function
        applied within the ``exclusion_radius``. The cutoff is computed as a raised cosine
        with exponent ``exclusion_degree``
    """

    def __init__(
        self,
        smearing: Optional[float] = None,
        exclusion_radius: Optional[float] = None,
        exclusion_degree: int = 1,
        epsilon: float = 0.0,
    ):
        super().__init__()

        self.exclusion_degree = exclusion_degree
        if smearing is not None:
            self.register_buffer(
                "smearing", torch.tensor(smearing, dtype=torch.float64)
            )
        else:
            self.smearing = None
        if exclusion_radius is not None:
            self.register_buffer(
                "exclusion_radius",
                torch.tensor(exclusion_radius, dtype=torch.float64),
            )
        else:
            self.exclusion_radius = None
        self.register_buffer("epsilon", torch.tensor(epsilon, dtype=torch.float64))

    @torch.jit.export
    def f_cutoff(self, vector: torch.Tensor) -> torch.Tensor:
        r"""
        Default cutoff function defining the *local* region that should be excluded from
        the computation of a long-range model. Defaults to a shifted cosine
        :math:`1 - ((1 - \cos \pi r/r_\mathrm{cut})/2) ^ n`. where :math:`n` is the
        ``exclusion_degree`` parameter.

        :param vector: torch.tensor containing the vectors at which the potential is to
            be evaluated.
        """
        r_mag = torch.norm(vector, dim=1, keepdim=True)
        if self.exclusion_radius is None:
            raise ValueError(
                "Cannot compute cutoff function when `exclusion_radius` is not set"
            )

        return torch.where(
            r_mag < self.exclusion_radius,
            1
            - ((1 - torch.cos(torch.pi * (r_mag / self.exclusion_radius))) * 0.5)
            ** self.exclusion_degree,
            0.0,
        )

    def from_dist(self, vector: torch.Tensor) -> torch.Tensor:
        r"""
        Full dipolar potential as a function of :math:`\mathbf{r}`.

        :param vector: torch.tensor containing the vectors at which the potential is to
            be evaluated.
        """
        r_mag = torch.norm(vector, dim=1, keepdim=True)
        scalar_potential = 1.0 / (r_mag**3)
        r_outer = torch.bmm(vector.unsqueeze(2), vector.unsqueeze(1))
        return scalar_potential.unsqueeze(-1) * torch.eye(3).to(r_outer).unsqueeze(
            0
        ) - 3.0 * r_outer / (r_mag**5).unsqueeze(-1)

    @torch.jit.export
    def sr_from_dist(self, vector: torch.Tensor) -> torch.Tensor:
        """
        Short-range part of the pair potential in real space.

        :param dist: torch.tensor containing the distance vectors at which the potential
            is to be evaluated.
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute range-separated potential when `smearing` is not specified."
            )
        if self.exclusion_radius is None:
            return self.from_dist(vector) - self.lr_from_dist(vector)
        return -self.lr_from_dist(vector) * self.f_cutoff(vector).unsqueeze(-1)

    @torch.jit.export
    def lr_from_dist(self, vector: torch.Tensor) -> torch.Tensor:
        r"""
        Long-range of the range-separated dipolar potential.

        Used to subtract out the interior contributions after computing the long-range
        part in reciprocal (Fourier) space.

        :param vector: torch.tensor containing the vectors at which the potential is to
            be evaluated.
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range contribution without specifying `smearing`."
            )
        alpha = 1 / (2 * self.smearing**2)
        r_mag = torch.norm(vector, dim=1, keepdim=True)
        r_outer = torch.bmm(vector.unsqueeze(2), vector.unsqueeze(1))
        B1 = torch.erfc(torch.sqrt(alpha) * r_mag) / r_mag**3
        B2 = 2 * torch.sqrt(alpha / torch.pi) * torch.exp(-alpha * r_mag**2) / r_mag**2
        B = 1.0 / (r_mag**3) - B1 - B2
        C1 = 3.0 * torch.erfc(torch.sqrt(alpha) * r_mag) / r_mag**5
        C2 = (
            2
            * torch.sqrt(alpha / torch.pi)
            * (2 * alpha + 3 / r_mag**2)
            * torch.exp(-alpha * r_mag**2)
            / r_mag**2
        )
        C = 3.0 / (r_mag**5) - C1 - C2
        return B.unsqueeze(-1) * torch.eye(3).to(r_outer).unsqueeze(
            0
        ) - r_outer * C.unsqueeze(-1)

    def lr_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        r"""
        Fourier transform of the long-range part of the potential.

        :param k_sq: torch.tensor containing the squared lengths (2-norms) of the wave
            vectors k at which the Fourier-transformed potential is to be evaluated
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range kernel without specifying `smearing`."
            )

        # avoid NaNs in backward, see
        # https://github.com/jax-ml/jax/issues/1052
        # https://github.com/tensorflow/probability/blob/main/discussion/where-nan.pdf
        masked = torch.where(k_sq == 0, 1.0, k_sq)
        return torch.where(
            k_sq == 0,
            0.0,
            4 * torch.pi * torch.exp(-0.5 * self.smearing**2 * masked) / masked,
        )

    def self_contribution(self) -> torch.Tensor:
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range contribution without specifying `smearing`."
            )
        alpha = 1 / (2 * self.smearing**2)
        return 4 * torch.pi / 3 * torch.sqrt((alpha / torch.pi) ** 3)

    def background_correction(self, volume) -> torch.Tensor:
        if self.epsilon == 0.0:
            return self.epsilon
        return 4 * torch.pi / (2 * self.epsilon + 1) / volume

    self_contribution.__doc__ = Potential.self_contribution.__doc__
    background_correction.__doc__ = Potential.background_correction.__doc__



import contextlib

from . import calculators, lib, potentials, prefactors, tuning  # noqa
from ._version import __version__, __version_tuple__  # noqa
from .calculators import (
    Calculator,
    CalculatorDipole,
    EwaldCalculator,
    P3MCalculator,
    PMECalculator,
)
from .potentials import (
    CombinedPotential,
    CoulombPotential,
    InversePowerLawPotential,
    Potential,
    PotentialDipole,
    SplinePotential,
)

with contextlib.suppress(ImportError):
    from . import metatensor  # noqa

__all__ = [
    "Calculator",
    "EwaldCalculator",
    "P3MCalculator",
    "PMECalculator",
    "CoulombPotential",
    "Potential",
    "InversePowerLawPotential",
    "SplinePotential",
    "CombinedPotential",
    "PotentialDipole",
    "CalculatorDipole",
]



from typing import Optional

import torch


class Potential(torch.nn.Module):
    r"""
    Base interface for a pair potential energy function between monopoles.

    The class provides the interface to compute a short-range and long-range functions
    in real space (such that :math:`V(r)=V_{\mathrm{SR}}(r)+V_{\mathrm{LR}}(r)` ), as
    well as a reciprocal-space version of the long-range component
    :math:`\hat{V}_{\mathrm{LR}}(k))` ).

    Derived classes can decide to implement a subset of these functionalities (e.g.
    providing only the real-space potential :math:`V(r)`). Internal state variables and
    parameters in derived classes should be defined in the ``__init__``  method.

    This base class also provides parameters to set the length scale associated with the
    range separation (``smearing``), and a cutoff function that can be optionally
    set to zero out the potential *inside* a short-range ``exclusion_radius``. This is
    often useful when combining ``torch-pme``-based ML models with local models that are
    better suited to describe the structure within a local cutoff.

    Note that a :class:`Potential` class can also be used inside a
    :class:`KSpaceFilter`, see :func:`Potential.kernel_from_k_sq`.

    :param smearing: The length scale associated with the switching between
        :math:`V_{\mathrm{SR}}(r)` and :math:`V_{\mathrm{LR}}(r)`
    :param exclusion_radius: A length scale that defines a *local environment* within
        which the potential should be smoothly zeroed out, as it will be described by a
        separate model.
    :param exclusion_degree: Controls the sharpness of the transition in the cutoff function
        applied within the ``exclusion_radius``. The cutoff is computed as a raised cosine
        with exponent ``exclusion_degree``
    """

    def __init__(
        self,
        smearing: Optional[float] = None,
        exclusion_radius: Optional[float] = None,
        exclusion_degree: int = 1,
    ):
        super().__init__()

        if smearing is not None:
            self.register_buffer(
                "smearing", torch.tensor(smearing, dtype=torch.float64)
            )
        else:
            self.smearing = None

        self.exclusion_radius = exclusion_radius
        self.exclusion_degree = exclusion_degree

    @torch.jit.export
    def f_cutoff(self, dist: torch.Tensor) -> torch.Tensor:
        r"""
        Default cutoff function defining the *local* region that should be excluded from
        the computation of a long-range model. Defaults to a shifted cosine
        :math:`1 - ((1 - \cos \pi r/r_\mathrm{cut})/2) ^ n`. where :math:`n` is the
        ``exclusion_degree`` parameter.

        :param dist: a torc.Tensor containing the interatomic distances over which the
            cutoff function should be computed.
        """
        if self.exclusion_radius is None:
            raise ValueError(
                "Cannot compute cutoff function when `exclusion_radius` is not set"
            )

        return torch.where(
            dist < self.exclusion_radius,
            1
            - ((1 - torch.cos(torch.pi * (dist / self.exclusion_radius))) * 0.5)
            ** self.exclusion_degree,
            0.0,
        )

    @torch.jit.export
    def from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        """
        Computes a pair potential given a tensor of interatomic distances.

        :param dist: torch.tensor containing the distances at which the potential
            is to be evaluated.
        """
        raise NotImplementedError(
            f"from_dist is not implemented for {self.__class__.__name__}"
        )

    @torch.jit.export
    def sr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        r"""
        Short-range (SR) part of the pair potential in real space.

        Even though one can provide a custom version, this is usually evaluated as
        :math:`V_{\mathrm{SR}}(r)=V(r)-V_{\mathrm{LR}}(r)`, based on the full and
        long-range parts of the potential. If the parameter ``exclusion_radius`` is
        defined, it computes this part as
        :math:`V_{\mathrm{SR}}(r)=-V_{\mathrm{LR}}(r)*f_\mathrm{cut}(r)` so that, when
        added to the part of the potential computed in the Fourier domain, the potential
        within the local region goes smoothly to zero.

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute range-separated potential when `smearing` is not specified."
            )
        if self.exclusion_radius is None:
            return self.from_dist(dist) - self.lr_from_dist(dist)
        return -self.lr_from_dist(dist) * self.f_cutoff(dist)

    @torch.jit.export
    def lr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        r"""
        Computes the long-range part of the pair potential :math:`V_\mathrm{LR}(r)`. in
        real space, given a tensor of interatomic distances.

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        raise NotImplementedError(
            f"lr_from_dist is not implemented for {self.__class__.__name__}"
        )

    @torch.jit.export
    def lr_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        r"""
        Computes the Fourier-domain version of the long-range part of the pair potential
        :math:`\hat{V}_\mathrm{LR}(k)`. The function is expressed in terms of
        :math:`k^2`, as that avoids, in several important cases, an unnecessary square
        root operation.
        :param k_sq: torch.tensor containing the squared norm of the Fourier domain
        vectors at which :math:`\hat{V}_\mathrm{LR}` must be evaluated.
        """
        raise NotImplementedError(
            f"lr_from_k_sq is not implemented for {self.__class__.__name__}"
        )

    @torch.jit.export
    def kernel_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        """
        Compatibility function with the interface of :class:`KSpaceKernel`, so that
        potentials can be used as kernels for :class:`KSpaceFilter`.
        """
        return self.lr_from_k_sq(k_sq)

    @torch.jit.export
    def self_contribution(self) -> torch.Tensor:
        """
        A correction that depends exclusively on the "charge" on every particle and on
        the range splitting parameter. Foe example, in the case of a Coulomb potential,
        this is the potential generated at the origin by the fictituous Gaussian charge
        density in order to split the potential into a SR and LR part.
        """
        raise NotImplementedError(
            f"self_contribution is not implemented for {self.__class__.__name__}"
        )

    @torch.jit.export
    def background_correction(self) -> torch.Tensor:
        """
        A correction designed to compensate for the presence of divergent terms. For
        instance, the energy of a periodic electrostatic system is infinite when the
        cell is not charge-neutral. This term then implicitly assumes that a homogeneous
        background charge of the opposite sign is present to make the cell neutral.
        """
        raise NotImplementedError(
            f"background_correction is not implemented for {self.__class__.__name__}"
        )



from .combined import CombinedPotential
from .coulomb import CoulombPotential
from .inversepowerlaw import InversePowerLawPotential
from .potential import Potential
from .potential_dipole import PotentialDipole
from .spline import SplinePotential

__all__ = [
    "CombinedPotential",
    "CoulombPotential",
    "InversePowerLawPotential",
    "Potential",
    "SplinePotential",
    "PotentialDipole",
]



from typing import Union

import torch


def _validate_parameters(
    charges: torch.Tensor,
    cell: torch.Tensor,
    positions: torch.Tensor,
    neighbor_indices: torch.Tensor,
    neighbor_distances: torch.Tensor,
    smearing: Union[float, None],
) -> None:
    dtype = positions.dtype
    device = positions.device

    # check shape, dtype and device of positions
    num_atoms = len(positions)
    if list(positions.shape) != [len(positions), 3]:
        raise ValueError(
            "`positions` must be a tensor with shape [n_atoms, 3], got tensor "
            f"with shape {list(positions.shape)}"
        )

    # check shape, dtype and device of cell
    if list(cell.shape) != [3, 3]:
        raise ValueError(
            "`cell` must be a tensor with shape [3, 3], got tensor with shape "
            f"{list(cell.shape)}"
        )

    if cell.dtype != dtype:
        raise TypeError(
            f"type of `cell` ({cell.dtype}) must be same as that of the `positions` class ({dtype})"
        )

    if cell.device != device:
        raise ValueError(
            f"device of `cell` ({cell.device}) must be same as that of the `positions` class ({device})"
        )

    if smearing is not None and torch.equal(
        cell.det(), torch.tensor(0.0, dtype=cell.dtype, device=cell.device)
    ):
        raise ValueError(
            "provided `cell` has a determinant of 0 and therefore is not valid for "
            "periodic calculation"
        )

    # check shape, dtype & device of `charges`
    if charges.dim() != 2:
        raise ValueError(
            "`charges` must be a 2-dimensional tensor, got "
            f"tensor with {charges.dim()} dimension(s) and shape "
            f"{list(charges.shape)}"
        )

    if list(charges.shape) != [num_atoms, charges.shape[1]]:
        raise ValueError(
            "`charges` must be a tensor with shape [n_atoms, n_channels], with "
            "`n_atoms` being the same as the variable `positions`. Got tensor with "
            f"shape {list(charges.shape)} where positions contains "
            f"{len(positions)} atoms"
        )

    if charges.dtype != dtype:
        raise TypeError(
            f"type of `charges` ({charges.dtype}) must be same as that of the `positions` class ({dtype})"
        )

    if charges.device != device:
        raise ValueError(
            f"device of `charges` ({charges.device}) must be same as that of the `positions` class "
            f"({device})"
        )

    # check shape, dtype & device of `neighbor_indices` and `neighbor_distances`
    if neighbor_indices.shape[1] != 2:
        raise ValueError(
            "neighbor_indices is expected to have shape [num_neighbors, 2]"
            f", but got {list(neighbor_indices.shape)} for one "
            "structure"
        )

    if neighbor_indices.device != device:
        raise ValueError(
            f"device of `neighbor_indices` ({neighbor_indices.device}) must be "
            f"same as that of the `positions` class ({device})"
        )

    if neighbor_distances.shape != neighbor_indices[:, 0].shape:
        raise ValueError(
            "`neighbor_indices` and `neighbor_distances` need to have shapes "
            "[num_neighbors, 2] and [num_neighbors], but got "
            f"{list(neighbor_indices.shape)} and {list(neighbor_distances.shape)}"
        )

    if neighbor_distances.device != device:
        raise ValueError(
            f"device of `neighbor_distances` ({neighbor_distances.device}) must be "
            f"same as that of the `positions` class ({device})"
        )

    if neighbor_distances.dtype != dtype:
        raise TypeError(
            f"type of `neighbor_distances` ({neighbor_distances.dtype}) must be same "
            f"as that of the `positions` class ({dtype})"
        )



import torch
from torch import profiler

from ..lib.kspace_filter import KSpaceFilter
from ..lib.kvectors import get_ns_mesh
from ..lib.mesh_interpolator import MeshInterpolator
from ..potentials import Potential
from .calculator import Calculator


class PMECalculator(Calculator):
    r"""
    Potential using a particle mesh-based Ewald (PME).

    Scaling as :math:`\mathcal{O}(NlogN)` with respect to the number of particles
    :math:`N` used as a reference to test faster implementations.

    For getting reasonable values for the ``smaring`` of the potential class and  the
    ``mesh_spacing`` based on a given accuracy for a specific structure you should use
    :func:`torchpme.tuning.tune_pme`. This function will also find the optimal
    ``cutoff`` for the  **neighborlist**.

    .. hint::

        For a training exercise it is recommended only run a tuning procedure with
        :func:`torchpme.tuning.tune_pme` for the largest system in your dataset.

    :param potential: A :class:`torchpme.potentials.Potential` object that implements
        the evaluation of short and long-range potential terms. The ``smearing``
        parameter of the potential determines the split between real and k-space
        regions. For a :class:`torchpme.CoulombPotential` it corresponds to the
        smearing of the atom-centered Gaussian used to split the Coulomb potential into
        the short- and long-range parts. A reasonable value for most systems is to set
        it to ``1/5`` times the neighbor list cutoff.
    :param mesh_spacing: Value that determines the umber of Fourier-space grid points
        that will be used along each axis. If set to None, it will automatically be set
        to half of ``smearing``.
    :param interpolation_nodes: The number ``n`` of nodes used in the interpolation per
        coordinate axis. The total number of interpolation nodes in 3D will be ``n^3``.
        In general, for ``n`` nodes, the interpolation will be performed by piecewise
        polynomials of degree ``n - 1`` (e.g. ``n = 4`` for cubic interpolation).
        Only the values ``3, 4, 5, 6, 7`` are supported.
    :param full_neighbor_list: If set to :obj:`True`, a "full" neighbor list
        is expected as input. This means that each atom pair appears twice. If
        set to :obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    """

    def __init__(
        self,
        potential: Potential,
        mesh_spacing: float,
        interpolation_nodes: int = 4,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
    ):
        super().__init__(
            potential=potential,
            full_neighbor_list=full_neighbor_list,
            prefactor=prefactor,
        )

        if potential.smearing is None:
            raise ValueError(
                "Must specify smearing to use a potential with PMECalculator"
            )

        self.mesh_spacing: float = mesh_spacing

        cell = torch.eye(
            3,
            device=self.potential.smearing.device,
            dtype=self.potential.smearing.dtype,
        )
        ns_mesh = torch.ones(3, dtype=int, device=cell.device)

        self.kspace_filter: KSpaceFilter = KSpaceFilter(
            cell=cell,
            ns_mesh=ns_mesh,
            kernel=self.potential,
            fft_norm="backward",
            ifft_norm="forward",
        )

        self.interpolation_nodes: int = interpolation_nodes

        self.mesh_interpolator: MeshInterpolator = MeshInterpolator(
            cell=cell,
            ns_mesh=ns_mesh,
            interpolation_nodes=self.interpolation_nodes,
            method="Lagrange",  # convention for classic PME
        )

    def _compute_kspace(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
    ) -> torch.Tensor:
        # TODO: Kernel function `G` and initialization of `MeshInterpolator` only depend
        # on `cell`. Caching may save up to 15% but issues with AD need to be resolved.

        with profiler.record_function("init 0: preparation"):
            # Compute number of times each basis vector of the reciprocal space can be
            # scaled until the cutoff is reached
            ns = get_ns_mesh(cell, self.mesh_spacing)

        with profiler.record_function("init 1: update mesh interpolator"):
            self.mesh_interpolator.update(cell, ns)

        with profiler.record_function("update the mesh for the k-space filter"):
            self.kspace_filter.update(cell, ns)

        with profiler.record_function("step 1: compute density interpolation"):
            self.mesh_interpolator.compute_weights(positions)
            rho_mesh = self.mesh_interpolator.points_to_mesh(particle_weights=charges)

        with profiler.record_function("step 2: perform actual convolution using FFT"):
            potential_mesh = self.kspace_filter.forward(rho_mesh)

        with profiler.record_function("step 3: back interpolation + volume scaling"):
            ivolume = torch.abs(cell.det()).pow(-1)
            interpolated_potential = (
                self.mesh_interpolator.mesh_to_points(potential_mesh) * ivolume
            )

        with profiler.record_function("step 4: remove the self-contribution"):
            # Using the Coulomb potential as an example, this is the potential generated
            # at the origin by the fictituous Gaussian charge density in order to split
            # the potential into a SR and LR part. This contribution always should be
            # subtracted since it depends on the smearing parameter, which is purely a
            # convergence parameter.
            interpolated_potential -= charges * self.potential.self_contribution()

        with profiler.record_function("step 5: charge neutralization"):
            # If the cell has a net charge (i.e. if sum(charges) != 0), the method
            # implicitly assumes that a homogeneous background charge of the opposite
            # sign is present to make the cell neutral. In this case, the potential has
            # to be adjusted to compensate for this. An extra factor of 2 is added to
            # compensate for the division by 2 later on
            charge_tot = torch.sum(charges, dim=0)
            prefac = self.potential.background_correction()
            interpolated_potential -= 2 * prefac * charge_tot * ivolume

        # Compensate for double counting of pairs (i,j) and (j,i)
        return interpolated_potential / 2



from .. import calculators as torch_calculators
from .calculator import Calculator


class PMECalculator(Calculator):
    r"""
    Potential using a particle mesh-based Ewald (PME).

    Refer to :class:`torchpme.PMECalculator` for parameter documentation.

    For an **example** on the usage for any calculator refer to :ref:`userdoc-how-to`.
    """

    # see torchpme.metatensor.calculator
    _base_calculator = torch_calculators.PMECalculator



from typing import Optional

import torch

from .potential import Potential


class CoulombPotential(Potential):
    """
    Smoothed electrostatic Coulomb potential :math:`1/r`.

    Here :math:`r` is the inter-particle distance

    It can be used to compute:

    1. the full :math:`1/r` potential
    2. its short-range (SR) and long-range (LR) parts, the split being determined by a
       length-scale parameter (called "Inverse" in the code)
    3. the Fourier transform of the LR part

    :param smearing: float or torch.Tensor containing the parameter often called "sigma"
        in publications, which determines the length-scale at which the short-range and
        long-range parts of the naive :math:`1/r` potential are separated. The smearing
        parameter corresponds to the "width" of a Gaussian smearing of the particle
        density.
    :param exclusion_radius: A length scale that defines a *local environment* within
        which the potential should be smoothly zeroed out, as it will be described by a
        separate model.
    :param exclusion_degree: Controls the sharpness of the transition in the cutoff function
        applied within the ``exclusion_radius``. The cutoff is computed as a raised cosine
        with exponent ``exclusion_degree``
    """

    def __init__(
        self,
        smearing: Optional[float] = None,
        exclusion_radius: Optional[float] = None,
        exclusion_degree: int = 1,
    ):
        super().__init__(smearing, exclusion_radius, exclusion_degree)

    def from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        """
        Full :math:`1/r` potential as a function of :math:`r`.

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        return 1.0 / dist

    def lr_from_dist(self, dist: torch.Tensor) -> torch.Tensor:
        """
        Long range of the range-separated :math:`1/r` potential.

        Used to subtract out the interior contributions after computing the LR part in
        reciprocal (Fourier) space.

        :param dist: torch.tensor containing the distances at which the potential is to
            be evaluated.
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range contribution without specifying `smearing`."
            )

        return torch.erf(dist / self.smearing / 2.0**0.5) / dist

    def lr_from_k_sq(self, k_sq: torch.Tensor) -> torch.Tensor:
        r"""
        Fourier transform of the LR part potential in terms of :math:`\mathbf{k^2}`.

        :param k_sq: torch.tensor containing the squared lengths (2-norms) of the wave
            vectors k at which the Fourier-transformed potential is to be evaluated
        """
        if self.smearing is None:
            raise ValueError(
                "Cannot compute long-range kernel without specifying `smearing`."
            )

        # avoid NaNs in backward, see
        # https://github.com/jax-ml/jax/issues/1052
        # https://github.com/tensorflow/probability/blob/main/discussion/where-nan.pdf
        masked = torch.where(k_sq == 0, 1.0, k_sq)
        return torch.where(
            k_sq == 0,
            0.0,
            4 * torch.pi * torch.exp(-0.5 * self.smearing**2 * masked) / masked,
        )

    def self_contribution(self) -> torch.Tensor:
        # self-correction for 1/r potential
        if self.smearing is None:
            raise ValueError(
                "Cannot compute self contribution without specifying `smearing`."
            )
        return (2 / torch.pi) ** 0.5 / self.smearing

    def background_correction(self) -> torch.Tensor:
        # "charge neutrality" correction for 1/r potential
        if self.smearing is None:
            raise ValueError(
                "Cannot compute background correction without specifying `smearing`."
            )
        return torch.pi * self.smearing**2

    self_contribution.__doc__ = Potential.self_contribution.__doc__
    background_correction.__doc__ = Potential.background_correction.__doc__



from .calculator import Calculator
from .ewald import EwaldCalculator
from .p3m import P3MCalculator
from .pme import PMECalculator

__all__ = ["Calculator", "EwaldCalculator", "P3MCalculator", "PMECalculator"]



from typing import Optional

import torch


class MeshInterpolator(torch.nn.Module):
    """
    Class for handling all steps related to interpolations in the context of a mesh
    based Ewald summation.

    In particular, this includes two core functionalities:
    1. "forwards" interpolation, in which the "charges" or more general
    "particle weights" of atoms are assigned to grid points of a mesh.
    This is done in the :func:`points_to_mesh` function.
    2. "backwards" interpolation, in which values defined
    on a mesh are interpolated to arbitrary positions typically lying between mesh
    points. This is done in the :func:`mesh_to_points` function.

    Since the computation of the interpolation weights for both of the above types
    of calculations is identical, this is performed in a separate function called
    :func:`compute_weights`.

    See also the :ref:`example-mesh-demo` for a demonstration of the
    functionalities of this class.

    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param ns_mesh: toch.tensor of shape ``(3,)``
        Number of mesh points to use along each of the three axes
    :param interpolation_nodes: int
        The number ``n`` of nodes used in the interpolation per
        coordinate axis. The total number of interpolation nodes in 3D will be ``n^3``.
        In general, for ``n`` nodes, the interpolation will be performed by piecewise
        polynomials of degree ``n - 1`` (e.g. ``n = 4`` for cubic interpolation).
        For Lagrange interpolation, only the values ``3, 4, 5, 6, 7`` are supported.
        For P3M interpolation, only the values ``1, 2, 3, 4, 5`` are supported.
    :param method: str
        The interpolation method to use. Either "Lagrange" or "P3M".
    """

    def __init__(
        self,
        cell: torch.Tensor,
        ns_mesh: torch.Tensor,
        interpolation_nodes: int,
        method: str,
    ):
        super().__init__()

        if method not in ["Lagrange", "P3M"]:
            raise ValueError(
                f"method '{method}' is not supported. Choose from 'Lagrange' or 'P3M'"
            )

        self.method: str = method
        self.interpolation_nodes: int = interpolation_nodes

        self.update(cell, ns_mesh)

        # TorchScript requires to initialize all attributes in __init__
        self.interpolation_weights: torch.Tensor = torch.zeros(
            1, device=self._device, dtype=self._dtype
        )
        self.x_shifts: torch.Tensor = torch.zeros(1, device=self._device)
        self.y_shifts: torch.Tensor = torch.zeros(1, device=self._device)
        self.z_shifts: torch.Tensor = torch.zeros(1, device=self._device)
        self.x_indices: torch.Tensor = torch.zeros(1, device=self._device)
        self.y_indices: torch.Tensor = torch.zeros(1, device=self._device)
        self.z_indices: torch.Tensor = torch.zeros(1, device=self._device)

    def update(
        self,
        cell: Optional[torch.Tensor] = None,
        ns_mesh: Optional[torch.Tensor] = None,
    ) -> None:
        """
        Update buffers and derived attributes of the instance.

        Call this to reuse a ``MeshInterpolator`` object when the ``cell`` parameters or
        the mesh resolution changes. If neither ``cell`` nor ``ns_mesh`` are passed
        there is nothing to be done.

        :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
            vector of the unit cell
        :param ns_mesh: toch.tensor of shape ``(3,)`` Number of mesh points to use along
            each of the three axes
        """
        if cell is not None:
            if cell.shape != (3, 3):
                raise ValueError(
                    f"cell of shape {list(cell.shape)} should be of shape (3, 3)"
                )
            self.cell = cell
            self.inverse_cell = cell.clone()
            self._dtype = cell.dtype
            self._device = cell.device

            if self.cell.is_cuda:
                # use function that does not synchronize with the CPU
                self.inverse_cell = torch.linalg.inv_ex(cell)[0]
            else:
                self.inverse_cell = torch.linalg.inv(cell)

        if ns_mesh is not None:
            if ns_mesh.shape != (3,):
                raise ValueError(
                    f"shape {list(ns_mesh.shape)} of `ns_mesh` has to be (3,)"
                )
            self.ns_mesh = ns_mesh

        if self.cell.device != self.ns_mesh.device:
            raise ValueError(
                "`cell` and `ns_mesh` are on different devices, got "
                f"{self.cell.device} and {self.ns_mesh.device}"
            )

    def get_mesh_xyz(self) -> torch.Tensor:
        """
        Returns the Cartesian positions of the mesh points.

        :return: torch.tensor of shape ``(nx, ny, nz, 3)``
            containing the positions of the grid points
        """
        nx = self.ns_mesh[0]
        ny = self.ns_mesh[1]
        nz = self.ns_mesh[2]

        grid_scaled = torch.stack(
            torch.meshgrid(
                torch.arange(nx, dtype=self._dtype, device=self._device) / nx,
                torch.arange(ny, dtype=self._dtype, device=self._device) / ny,
                torch.arange(nz, dtype=self._dtype, device=self._device) / nz,
                indexing="ij",
            ),
            dim=-1,
        )
        return torch.matmul(grid_scaled, self.cell)

    def _compute_1d_weights(self, x: torch.Tensor) -> torch.Tensor:
        if self.method == "Lagrange":
            return self._compute_1d_weights_Lagrange(x)
        if self.method == "P3M":
            return self._compute_1d_weights_P3M(x)
        raise ValueError("Only `method` `Lagrange` and `P3M` are allowed")

    def _compute_1d_weights_P3M(self, x: torch.Tensor) -> torch.Tensor:
        """
        Generate the smooth interpolation weights used to smear the particles onto a
        mesh.

        The details of the method are described in
        `J. Chem. Phys. 109, 7678–7693 (1998) <https://doi.org/10.1063/1.477414>`_

        :param x: torch.tensor of shape ``(n,)``
            Set of relative positions in the interval [-1/2, 1/2].

        :return: torch.tensor of shape ``(interpolation_nodes, n)``
            Interpolation weights
        """
        # Compute weights based on the given interpolation_nodes
        if self.interpolation_nodes == 1:
            return torch.ones(
                (1, x.shape[0], x.shape[1]), dtype=self._dtype, device=self._device
            )
        if self.interpolation_nodes == 2:
            return torch.stack([0.5 * (1 - 2 * x), 0.5 * (1 + 2 * x)])

        x2 = x * x
        if self.interpolation_nodes == 3:
            return torch.stack(
                [
                    1 / 8 * (1 - 4 * x + 4 * x2),
                    1 / 4 * (3 - 4 * x2),
                    1 / 8 * (1 + 4 * x + 4 * x2),
                ]
            )

        x3 = x * x2
        if self.interpolation_nodes == 4:
            return torch.stack(
                [
                    1 / 48 * (1 - 6 * x + 12 * x2 - 8 * x3),
                    1 / 48 * (23 - 30 * x - 12 * x2 + 24 * x3),
                    1 / 48 * (23 + 30 * x - 12 * x2 - 24 * x3),
                    1 / 48 * (1 + 6 * x + 12 * x2 + 8 * x3),
                ]
            )

        x4 = x * x3
        if self.interpolation_nodes == 5:
            return torch.stack(
                [
                    1 / 384 * (1 - 8 * x + 24 * x2 - 32 * x3 + 16 * x4),
                    1 / 96 * (19 - 44 * x + 24 * x2 + 16 * x3 - 16 * x4),
                    1 / 192 * (115 - 120 * x2 + 48 * x4),
                    1 / 96 * (19 + 44 * x + 24 * x2 - 16 * x3 - 16 * x4),
                    1 / 384 * (1 + 8 * x + 24 * x2 + 32 * x3 + 16 * x4),
                ]
            )
        raise ValueError("Only `interpolation_nodes` from 1 to 5 are allowed")

    def _compute_1d_weights_Lagrange(self, x: torch.Tensor) -> torch.Tensor:
        """
        Generate the smooth interpolation weights used to smear the particles onto a
        mesh.

        The details of the method are described in
        `J. Chem. Phys. 103, 3668-3679 (1995) <https://doi.org/10.1063/1.470043>`_

        :param x: torch.tensor of shape ``(n,)``
            Set of relative positions in the interval [-1/2, 1/2].

        :return: torch.tensor of shape ``(interpolation_nodes, n)``
            Interpolation weights
        """
        # Compute weights based on the given interpolation_nodes
        x2 = x * x
        if self.interpolation_nodes == 3:
            return torch.stack(
                [
                    1 / 2 * (-x + x2),
                    1 / 2 * (2 - 2 * x2),
                    1 / 2 * (x + x2),
                ]
            )

        x3 = x * x2
        if self.interpolation_nodes == 4:
            return torch.stack(
                [
                    1 / 48 * (-3 + 2 * x + 12 * x2 - 8 * x3),
                    1 / 48 * (27 - 54 * x - 12 * x2 + 24 * x3),
                    1 / 48 * (27 + 54 * x - 12 * x2 - 24 * x3),
                    1 / 48 * (-3 - 2 * x + 12 * x2 + 8 * x3),
                ]
            )

        x4 = x * x3
        if self.interpolation_nodes == 5:
            return torch.stack(
                [
                    1 / 24 * (2 * x - x2 - 2 * x3 + x4),
                    1 / 24 * (-16 * x + 16 * x2 + 4 * x3 - 4 * x4),
                    1 / 24 * (24 - 30 * x2 + 6 * x4),
                    1 / 24 * (16 * x + 16 * x2 - 4 * x3 - 4 * x4),
                    1 / 24 * (-2 * x - x2 + 2 * x3 + x4),
                ]
            )

        x5 = x * x4
        if self.interpolation_nodes == 6:
            return torch.stack(
                [
                    1 / 3840 * (45 - 18 * x - 200 * x2 + 80 * x3 + 80 * x4 - 32 * x5),
                    1
                    / 3840
                    * (-375 + 250 * x + 1560 * x2 - 1040 * x3 - 240 * x4 + 160 * x5),
                    1
                    / 3840
                    * (2250 - 4500 * x - 1360 * x2 + 2720 * x3 + 160 * x4 - 320 * x5),
                    1
                    / 3840
                    * (2250 + 4500 * x - 1360 * x2 - 2720 * x3 + 160 * x4 + 320 * x5),
                    1
                    / 3840
                    * (-375 - 250 * x + 1560 * x2 + 1040 * x3 - 240 * x4 - 160 * x5),
                    1 / 3840 * (45 + 18 * x - 200 * x2 - 80 * x3 + 80 * x4 + 32 * x5),
                ]
            )
        x6 = x * x5
        if self.interpolation_nodes == 7:
            return torch.stack(
                [
                    1 / 720 * (-12 * x + 4 * x2 + 15 * x3 - 5 * x4 - 3 * x5 + x6),
                    1
                    / 720
                    * (108 * x - 54 * x2 - 120 * x3 + 60 * x4 + 12 * x5 - 6 * x6),
                    1
                    / 720
                    * (-540 * x + 540 * x2 + 195 * x3 - 195 * x4 - 15 * x5 + 15 * x6),
                    1 / 720 * (720 - 980 * x2 + 280 * x4 - 20 * x6),
                    1
                    / 720
                    * (540 * x + 540 * x2 - 195 * x3 - 195 * x4 + 15 * x5 + 15 * x6),
                    1
                    / 720
                    * (-108 * x - 54 * x2 + 120 * x3 + 60 * x4 - 12 * x5 - 6 * x6),
                    1 / 720 * (12 * x + 4 * x2 - 15 * x3 - 5 * x4 + 3 * x5 + x6),
                ]
            )
        raise ValueError("Only `interpolation_nodes` from 3 to 7 are allowed")

    def compute_weights(self, positions: torch.Tensor):
        """
        Compute the interpolation weights of each atom for a given cell (specified
        during initialization of this class). The weights are not returned, but are used
        when calling the forward (:func:`points_to_mesh`) and backward
        (:func:`mesh_to_points`) interpolation functions.

        :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
            coordinates of the ``N`` particles within the supercell.
        """
        if positions.device != self._device:
            raise ValueError(
                f"`positions` device {positions.device} is not the same as instance "
                f"device {self._device}"
            )

        n_positions = len(positions)
        if positions.shape != (n_positions, 3):
            raise ValueError(
                f"shape {list(positions.shape)} of `positions` has to be (N, 3)"
            )

        # Compute positions relative to the mesh basis vectors
        positions_rel = self.ns_mesh * torch.matmul(positions, self.inverse_cell)

        # Calculate positions and distances based on interpolation nodes
        even = self.interpolation_nodes % 2 == 0
        if even:
            # For Lagrange interpolation, when the order is odd, the relative position
            # of a charge is the midpoint of the two nearest gridpoints. For P3M, the
            # same is true for even orders.
            positions_rel_idx = torch.floor(positions_rel).long()
            offsets = positions_rel - (positions_rel_idx + 1 / 2)
        else:
            # For Lagrange interpolation, when the order is even, the relative position
            # of a charge is the nearest gridpoint. For P3M, the same is true for
            # odd orders.
            positions_rel_idx = torch.round(positions_rel).long()
            offsets = positions_rel - positions_rel_idx

        # Compute weights based on distances and number of nodes
        self.interpolation_weights = self._compute_1d_weights(offsets)

        # Calculate indices of mesh points on which the particle weights are
        # interpolated. For each particle, its weight is "smeared" onto
        # `interpolation_nodes**3` mesh points, which can be achived using meshgrid
        # below.
        indices_to_interpolate = torch.stack(
            [
                (positions_rel_idx + i) % self.ns_mesh
                for i in range(
                    1 - (self.interpolation_nodes + 1) // 2,
                    1 + self.interpolation_nodes // 2,
                )
            ],
            dim=0,
        )

        # Generate shifts for x, y, z axes and flatten for indexing
        x_shifts, y_shifts, z_shifts = torch.meshgrid(
            torch.arange(self.interpolation_nodes, device=self._device),
            torch.arange(self.interpolation_nodes, device=self._device),
            torch.arange(self.interpolation_nodes, device=self._device),
            indexing="ij",
        )
        self.x_shifts = x_shifts.flatten()
        self.y_shifts = y_shifts.flatten()
        self.z_shifts = z_shifts.flatten()

        # Generate a flattened representation of all the indices
        # of the mesh points on which we wish to interpolate the
        # density.
        self.x_indices = indices_to_interpolate[self.x_shifts, :, 0]
        self.y_indices = indices_to_interpolate[self.y_shifts, :, 1]
        self.z_indices = indices_to_interpolate[self.z_shifts, :, 2]

    def points_to_mesh(self, particle_weights: torch.Tensor) -> torch.Tensor:
        """
        Generate a discretized density from interpolation weights. It assumes that
        :func:`compute_weights` has been called before to compute all the necessary
        weights and indices.

        :param particle_weights: torch.tensor of shape ``(n_points, n_channels)``
            ``particle_weights[i,a]`` is the weight (charge) that point (atom) i has to
            generate the "a-th" potential. In practice, this can be used to compute e.g.
            the Na and Cl contributions to the potential separately by using a one-hot
            encoding of the types.

        :return: torch.tensor of shape ``(n_channels, n_mesh, n_mesh, n_mesh)``
            Discrete density
        """
        if particle_weights.device != self._device:
            raise ValueError(
                f"`particle_weights` device {particle_weights.device} is not the same "
                f"as instance device {self._device}"
            )

        if particle_weights.dim() != 2:
            raise ValueError(
                f"`particle_weights` of dimension {particle_weights.dim()} has to be "
                "of dimension 2"
            )

        # Update mesh values by combining particle weights and interpolation weights
        n_channels = particle_weights.shape[1]
        nx = int(self.ns_mesh[0])
        ny = int(self.ns_mesh[1])
        nz = int(self.ns_mesh[2])
        rho_mesh = torch.zeros(
            (n_channels, nx, ny, nz), dtype=self._dtype, device=self._device
        )
        for a in range(n_channels):
            rho_mesh[a].index_put_(
                (self.x_indices, self.y_indices, self.z_indices),
                (
                    particle_weights[:, a]
                    * self.interpolation_weights[self.x_shifts, :, 0]
                    * self.interpolation_weights[self.y_shifts, :, 1]
                    * self.interpolation_weights[self.z_shifts, :, 2]
                ),
                accumulate=True,
            )

        return rho_mesh

    def mesh_to_points(self, mesh_vals: torch.Tensor) -> torch.Tensor:
        """
        Take a function defined on a mesh and interpolate
        its values on arbitrary positions.

        :param mesh_vals: torch.tensor of shape ``(n_channels, nx, ny, nz)``
            The tensor contains the values of a function evaluated on a
            three-dimensional mesh. ``(nx, ny, nz)`` are the number of
            points along each of the three directions, while ``n_channels``
            provides the number of such functions
            that are treated simulateously for the present system.

        :return: interpolated_values: torch.tensor of shape ``(n_points, n_channels)``
            Values of the interpolated function.
        """
        if mesh_vals.dim() != 4:
            raise ValueError(
                f"`mesh_vals` of dimension {mesh_vals.dim()} has to be of dimension 4"
            )

        return (
            (
                mesh_vals[:, self.x_indices, self.y_indices, self.z_indices]
                * self.interpolation_weights[self.x_shifts, :, 0]
                * self.interpolation_weights[self.y_shifts, :, 1]
                * self.interpolation_weights[self.z_shifts, :, 2]
            )
            .sum(dim=1)
            .T
        )



import math
from itertools import product
from typing import Any
from warnings import warn

import torch

from ..calculators import P3MCalculator
from .tuner import GridSearchTuner, TuningErrorBounds

# Coefficients for the P3M Fourier error,
# see Table II of http://dx.doi.org/10.1063/1.477415
A_COEF = [
    [None, 2 / 3, 1 / 50, 1 / 588, 1 / 4320, 1 / 23_232, 691 / 68_140_800, 1 / 345_600],
    [
        None,
        None,
        5 / 294,
        7 / 1440,
        3 / 1936,
        7601 / 13_628_160,
        13 / 57_600,
        3617 / 35_512_320,
    ],
    [
        None,
        None,
        None,
        21 / 3872,
        7601 / 2_271_360,
        143 / 69_120,
        47_021 / 35_512_320,
        745_739 / 838_397_952,
    ],
    [
        None,
        None,
        None,
        None,
        143 / 28_800,
        517_231 / 106_536_960,
        9_694_607 / 2_095_994_880,
        56_399_353 / 12_773_376_000,
    ],
    [
        None,
        None,
        None,
        None,
        None,
        106_640_677 / 11_737_571_328,
        733_191_589 / 59_609_088_000,
        25_091_609 / 1_560_084_480,
    ],
    [
        None,
        None,
        None,
        None,
        None,
        None,
        326_190_917 / 11_700_633_600,
        1_755_948_832_039 / 36_229_939_200_000,
    ],
    [None, None, None, None, None, None, None, 4_887_769_399 / 37_838_389_248],
]


def tune_p3m(
    charges: torch.Tensor,
    cell: torch.Tensor,
    positions: torch.Tensor,
    cutoff: float,
    neighbor_indices: torch.Tensor,
    neighbor_distances: torch.Tensor,
    full_neighbor_list: bool = False,
    prefactor: float = 1.0,
    exponent: int = 1,
    nodes_lo: int = 2,
    nodes_hi: int = 5,
    mesh_lo: int = 2,
    mesh_hi: int = 7,
    accuracy: float = 1e-3,
) -> tuple[float, dict[str, Any], float]:
    r"""
    Find the optimal parameters for :class:`torchpme.calculators.pme.PMECalculator`.

    For the error formulas are given `here <https://doi.org/10.1063/1.477415>`_. Note
    the difference notation between the parameters in the reference and ours:

    .. math::

        \alpha = \left(\sqrt{2}\,\mathrm{smearing} \right)^{-1}

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.
    :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
        which the potential should be computed in real space.
    :param neighbor_distances: torch.tensor with the pair distances of the neighbors for
        which the potential should be computed in real space.
    :param full_neighbor_list: If set to :py:obj:`True`, a "full" neighbor list
        is expected as input. This means that each atom pair appears twice. If
        set to :py:obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    :param cutoff: float, cutoff distance for the neighborlist supported
    :param exponent: :math:`p` in :math:`1/r^p` potentials, currently only :math:`p=1`
        is
    :param nodes_lo: Minimum number of interpolation nodes
    :param nodes_hi: Maximum number of interpolation nodes
    :param mesh_lo: Controls the minimum number of mesh points along the shortest axis,
        :math:`2^{mesh_lo}`
    :param mesh_hi: Controls the maximum number of mesh points along the shortest axis,
        :math:`2^{mesh_hi}`
    :param accuracy: Recomended values for a balance between the accuracy and speed is
        :math:`10^{-3}`. For more accurate results, use :math:`10^{-6}`.

    :return: Tuple containing a float of the optimal smearing for the :py:class:
        `CoulombPotential`, a dictionary with the parameters for
        :py:class:`P3MCalculator` and a float of the optimal cutoff value for the
        neighborlist computation, and the timing of this set of parameters.

    Example
    -------
    >>> import torch

    To allow reproducibility, we set the seed to a fixed value

    >>> _ = torch.manual_seed(0)
    >>> positions = torch.tensor([[0.0, 0.0, 0.0], [0.4, 0.4, 0.4]])
    >>> charges = torch.tensor([[1.0], [-1.0]])
    >>> cell = torch.eye(3)
    >>> neighbor_distances = torch.tensor(
    ...     [0.9381, 0.9381, 0.8246, 0.9381, 0.8246, 0.8246, 0.6928],
    ... )
    >>> neighbor_indices = torch.tensor(
    ...     [[0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1], [0, 1]]
    ... )
    >>> smearing, parameter, timing = tune_p3m(
    ...     charges,
    ...     cell,
    ...     positions,
    ...     cutoff=1.0,
    ...     neighbor_distances=neighbor_distances,
    ...     neighbor_indices=neighbor_indices,
    ...     accuracy=1e-1,
    ... )

    """
    # if cell is 0 `min_dimension` will be zero as well; we raise a propper error later
    min_dimension = float(torch.min(torch.linalg.norm(cell, dim=1)))
    params = [
        {
            "interpolation_nodes": interpolation_nodes,
            "mesh_spacing": 2 * min_dimension / (2**ns - 1),
        }
        for interpolation_nodes, ns in product(
            range(nodes_lo, nodes_hi + 1), range(mesh_lo, mesh_hi + 1)
        )
    ]

    tuner = GridSearchTuner(
        charges=charges,
        cell=cell,
        positions=positions,
        cutoff=cutoff,
        exponent=exponent,
        neighbor_indices=neighbor_indices,
        neighbor_distances=neighbor_distances,
        full_neighbor_list=full_neighbor_list,
        prefactor=prefactor,
        calculator=P3MCalculator,
        error_bounds=P3MErrorBounds(charges=charges, cell=cell, positions=positions),
        params=params,
    )
    smearing = tuner.estimate_smearing(accuracy)
    errs, timings = tuner.tune(accuracy)

    # There are multiple errors below the accuracy, return the one with the shortest
    # calculation time. The timing of those parameters leading to an higher error
    # than the accuracy are set to infinity
    if any(err < accuracy for err in errs):
        return smearing, params[timings.index(min(timings))], min(timings)
    # No parameter meets the requirement, return the one with the smallest error, and
    # throw a warning
    warn(
        f"No parameter meets the accuracy requirement.\n"
        f"Returning the parameter with the smallest error, which is {min(errs)}.\n",
        stacklevel=1,
    )
    return smearing, params[errs.index(min(errs))], timings[errs.index(min(errs))]


class P3MErrorBounds(TuningErrorBounds):
    r"""
    " Error bounds for :class:`torchpme.calculators.pme.P3MCalculator`.

    .. note::

        The :func:`torchpme.tuning.p3m.P3MErrorBounds.forward` method takes floats as
        the input, in order to be in consistency with the rest of the package -- these
        parameters are always ``float`` but not ``torch.Tensor``. This design, however,
        prevents the utilization of ``torch.autograd`` and other ``torch`` features. To
        take advantage of these features, one can use the
        :func:`torchpme.tuning.p3m.P3MErrorBounds.err_rspace` and
        :func:`torchpme.tuning.p3m.P3MErrorBounds.err_kspace`, which takes
        ``torch.Tensor`` as parameters.

    :param charges: torch.tensor of shape ``(len(positions, 1))`` containing the atomic
        (pseudo-)charges
    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
        coordinates of the ``N`` particles within the supercell.

    Example
    -------
    >>> import torch
    >>> positions = torch.tensor(
    ...     [[0.0, 0.0, 0.0], [0.4, 0.4, 0.4]], dtype=torch.float64
    ... )
    >>> charges = torch.tensor([[1.0], [-1.0]], dtype=torch.float64)
    >>> cell = torch.eye(3, dtype=torch.float64)
    >>> error_bounds = P3MErrorBounds(charges, cell, positions)
    >>> print(
    ...     error_bounds(
    ...         smearing=1.0, mesh_spacing=0.5, cutoff=4.4, interpolation_nodes=3
    ...     )
    ... )
    tensor(0.0005, dtype=torch.float64)

    """

    def __init__(
        self, charges: torch.Tensor, cell: torch.Tensor, positions: torch.Tensor
    ):
        super().__init__(charges, cell, positions)

        self.volume = torch.abs(torch.det(cell))
        self.sum_squared_charges = (charges**2).sum()
        self.prefac = 2 * self.sum_squared_charges / math.sqrt(len(positions))
        self.cell_dimensions = torch.linalg.norm(cell, dim=1)
        self.cell = cell
        self.positions = positions

    def err_kspace(
        self,
        smearing: torch.Tensor,
        mesh_spacing: torch.Tensor,
        interpolation_nodes: torch.Tensor,
    ) -> torch.Tensor:
        """
        The Fourier space error of P3M.

        :param smearing: see :class:`torchpme.P3MCalculator` for details
        :param mesh_spacing: see :class:`torchpme.P3MCalculator` for details
        :param interpolation_nodes: see :class:`torchpme.P3MCalculator` for details
        """
        actual_spacing = self.cell_dimensions / (
            2 * self.cell_dimensions / mesh_spacing + 1
        )
        h = torch.prod(actual_spacing) ** (1 / 3)

        return (
            self.prefac
            / self.volume ** (2 / 3)
            * (h * (1 / 2**0.5 / smearing)) ** interpolation_nodes
            * torch.sqrt(
                (1 / 2**0.5 / smearing)
                * self.volume ** (1 / 3)
                * math.sqrt(2 * torch.pi)
                * sum(
                    A_COEF[m][interpolation_nodes]
                    * (h * (1 / 2**0.5 / smearing)) ** (2 * m)
                    for m in range(interpolation_nodes)
                )
            )
        )

    def err_rspace(self, smearing: torch.Tensor, cutoff: torch.Tensor) -> torch.Tensor:
        """
        The real space error of P3M.

        :param smearing: see :class:`torchpme.P3MCalculator` for details
        :param cutoff: see :class:`torchpme.P3MCalculator` for details
        """
        return (
            self.prefac
            / torch.sqrt(cutoff * self.volume)
            * torch.exp(-(cutoff**2) / 2 / smearing**2)
        )

    def forward(
        self,
        smearing: float,
        mesh_spacing: float,
        cutoff: float,
        interpolation_nodes: int,
    ) -> torch.Tensor:
        r"""
        Calculate the error bound of P3M.

        .. math::
            \text{Error}_{\text{total}} = \sqrt{\text{Error}_{\text{real space}}^2 +
            \text{Error}_{\text{Fourier space}}^2

        :param smearing: see :class:`torchpme.P3MCalculator` for details
        :param mesh_spacing: see :class:`torchpme.P3MCalculator` for details
        :param cutoff: see :class:`torchpme.P3MCalculator` for details
        :param interpolation_nodes: see :class:`torchpme.P3MCalculator` for details
        """
        smearing = torch.tensor(smearing)
        mesh_spacing = torch.tensor(mesh_spacing)
        cutoff = torch.tensor(cutoff)
        interpolation_nodes = torch.tensor(interpolation_nodes)
        return torch.sqrt(
            self.err_kspace(smearing, mesh_spacing, interpolation_nodes) ** 2
            + self.err_rspace(smearing, cutoff) ** 2
        )



from .. import calculators as torch_calculators
from .calculator import Calculator


class P3MCalculator(Calculator):
    r"""
    Potential using a particle-particle particle-mesh based Ewald (P3M).

    Refer to :class:`torchpme.P3MCalculator` for parameter documentation.

    For an **example** on the usage for any calculator refer to :ref:`userdoc-how-to`.
    """

    # see torchpme.metatensor.calculator
    _base_calculator = torch_calculators.P3MCalculator



from typing import Optional

import torch
from torch import profiler

from .._utils import _validate_parameters
from ..lib import generate_kvectors_for_ewald
from ..potentials import PotentialDipole


class CalculatorDipole(torch.nn.Module):
    """
    Base calculator for interacting dipoles in the torch interface.

    :param potential: a :class:`PotentialDipole` class object containing the functions
        that are necessary to compute the various components of the potential, as
        well as the parameters that determine the behavior of the potential itself.
    :param full_neighbor_list: parameter indicating whether the neighbor information
        will come from a full (True) or half (False, default) neighbor list.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    :param lr_wavelength: the wavelength of the long-range part of the potential.
    """

    def __init__(
        self,
        potential: PotentialDipole,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
        lr_wavelength: Optional[float] = None,
    ):
        super().__init__()

        if not isinstance(potential, PotentialDipole):
            raise TypeError(
                f"Potential must be an instance of PotentialDipole, got {type(potential)}"
            )

        self.potential = potential
        self.lr_wavelength = lr_wavelength

        assert (
            self.lr_wavelength is not None
            and self.potential.smearing is not None
            or (self.lr_wavelength is None and self.potential.smearing is None)
        ), "Either both `lr_wavelength` and `smearing` must be set or both must be None"

        self.full_neighbor_list = full_neighbor_list

        self.prefactor = prefactor

    def _compute_rspace(
        self,
        dipoles: torch.Tensor,
        neighbor_indices: torch.Tensor,
        neighbor_vectors: torch.Tensor,
    ) -> torch.Tensor:
        # Compute the pair potential terms V(r_ij) for each pair of atoms (i,j)
        # contained in the neighbor list
        with profiler.record_function("compute bare potential"):
            if self.potential.smearing is None:
                potentials_bare = self.potential.from_dist(neighbor_vectors)
            else:
                potentials_bare = self.potential.sr_from_dist(neighbor_vectors)

        # Multiply the bare potential terms V(r_ij) with the corresponding dipoles
        # of ``atom j'' to obtain q_j*V(r_ij). Since each atom j can be a neighbor of
        # multiple atom i's, we need to access those from neighbor_indices
        atom_is = neighbor_indices[:, 0]
        atom_js = neighbor_indices[:, 1]
        with profiler.record_function("compute real potential"):
            contributions_is = torch.bmm(
                potentials_bare, dipoles[atom_js].unsqueeze(-1)
            ).squeeze(-1)

        # For each atom i, add up all contributions of the form q_j*V(r_ij) for j
        # ranging over all of its neighbors.
        with profiler.record_function("assign potential"):
            potential = torch.zeros_like(dipoles)
            potential.index_add_(0, atom_is, contributions_is)
            # If we are using a half neighbor list, we need to add the contributions
            # from the "inverse" pairs (j, i) to the atoms i
            if not self.full_neighbor_list:
                contributions_js = torch.bmm(
                    potentials_bare, dipoles[atom_is].unsqueeze(-1)
                ).squeeze(-1)
                potential.index_add_(0, atom_js, contributions_js)

        # Compensate for double counting of pairs (i,j) and (j,i)
        return potential / 2

    def _compute_kspace(
        self,
        dipoles: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
    ) -> torch.Tensor:
        # Define k-space cutoff from required real-space resolution
        k_cutoff = 2 * torch.pi / self.lr_wavelength

        # Compute number of times each basis vector of the reciprocal space can be
        # scaled until the cutoff is reached
        basis_norms = torch.linalg.norm(cell, dim=1)
        ns_float = k_cutoff * basis_norms / 2 / torch.pi
        ns = torch.ceil(ns_float).long()

        # Generate k-vectors and evaluate
        kvectors = generate_kvectors_for_ewald(ns=ns, cell=cell)
        knorm_sq = torch.sum(kvectors**2, dim=1)
        # We remove the singularity at k=0 by explicitly setting its
        # value to be equal to zero. This mathematically corresponds
        # to the requirement that the net charge of the cell is zero.
        # G = 4 * torch.pi * torch.exp(-0.5 * smearing**2 * knorm_sq) / knorm_sq
        G = self.potential.lr_from_k_sq(knorm_sq)

        # Compute the energy using the explicit method that
        # follows directly from the Poisson summation formula.
        # For this, we precompute trigonometric factors for optimization, which leads
        # to N^2 rather than N^3 scaling.
        trig_args = kvectors @ (positions.T)  # [k, i]
        c = torch.cos(trig_args)  # [k, i]
        s = torch.sin(trig_args)  # [k, i]
        sc = torch.stack([c, s], dim=0)  # [2 "f", k, i]
        mu_k = dipoles @ kvectors.T  # [i, k]
        sc_summed_G = torch.einsum("fki, ik, k->fk", sc, mu_k, G)
        energy = torch.einsum("fk, fki, kc->ic", sc_summed_G, sc, kvectors)
        energy /= torch.abs(cell.det())
        energy -= dipoles * self.potential.self_contribution()
        energy += self.potential.background_correction(
            torch.abs(cell.det())
        ) * dipoles.sum(dim=0)
        return energy / 2

    def forward(
        self,
        dipoles: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
        neighbor_indices: torch.Tensor,
        neighbor_vectors: torch.Tensor,
    ):
        r"""
        Compute the potential "energy".

        It is calculated as:

        .. math::

            V_i = \frac{1}{2} \sum_{j} \boldsymbol{\mu_j} \, \mathbf{v}(\mathbf{r_{ij}})

        where :math:`\mathbf{v}(\mathbf{r})` is the pair potential defined by the ``potential``
        parameter, and :math:`\boldsymbol{\mu_j}` are atomic "dipoles".

        If the ``smearing`` of the ``potential`` is not set, the calculator evaluates
        only the real-space part of the potential. Otherwise, provided that the
        calculator implements a ``_compute_kspace`` method, it will also evaluate the
        long-range part using a Fourier-domain method.

        :param dipoles: torch.tensor of shape ``(len(positions), 3)``
            containaing the atomic dipoles.
        :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
            vector of the unit cell
        :param positions: torch.tensor of shape ``(N, 3)`` containing the Cartesian
            coordinates of the ``N`` particles within the supercell.
        :param neighbor_indices: torch.tensor with the ``i,j`` indices of neighbors for
            which the potential should be computed in real space.
        :param neighbor_vectors: torch.tensor with the pair vectors of the neighbors
            for which the potential should be computed in real space.
        """
        # TODO: _validate_parameters to allow also dipoles. Temporarily pass the
        # distance tensor.
        _validate_parameters(
            charges=dipoles,
            cell=cell,
            positions=positions,
            neighbor_indices=neighbor_indices,
            neighbor_distances=neighbor_vectors.norm(dim=-1),
            smearing=self.potential.smearing,
        )

        # Compute short-range (SR) part using a real space sum
        potential_sr = self._compute_rspace(
            dipoles=dipoles,
            neighbor_indices=neighbor_indices,
            neighbor_vectors=neighbor_vectors,
        )

        if self.potential.smearing is None:
            return self.prefactor * potential_sr
        # Compute long-range (LR) part using a Fourier / reciprocal space sum
        potential_lr = self._compute_kspace(
            dipoles=dipoles,
            cell=cell,
            positions=positions,
        )

        return self.prefactor * (potential_sr + potential_lr)



#: Conversion factor from Gaussian units to SI units
SI = 2.3070775523417355e-28

#: Conversion factor from Gaussian units to electron volts / Angstroms
eV_A = 14.399645478425667

#: Conversion factor from Gaussian units to kilocalories per mole / Angstroms
kcalmol_A = 332.0637132991921

#: Conversion factor from Gaussian units to kilojoules per mole / Angstroms
kJmol = 1389.3545764438197



import torch

from ..lib import generate_kvectors_for_ewald
from ..potentials import Potential
from .calculator import Calculator


class EwaldCalculator(Calculator):
    r"""
    Potential computed using the Ewald sum.

    Scaling as :math:`\mathcal{O}(N^2)` with respect to the number of particles
    :math:`N`.

    For getting reasonable values for the ``smaring`` of the potential class and  the
    ``lr_wavelength`` based on a given accuracy for a specific structure you should use
    :func:`torchpme.tuning.tune_ewald`. This function will also find the optimal
    ``cutoff`` for the  **neighborlist**.

    .. hint::

        For a training exercise it is recommended only run a tuning procedure with
        :func:`torchpme.tuning.tune_ewald` for the largest system in your dataset.

    For a :math:`\mathcal{O}(N^{1.5})` scaling, one can set the parameters as following:

    .. math::

        \mathrm{smearing} &= 1.3 \cdot N^{1 / 6} / \sqrt{2}

        \mathrm{lr\_wavelength} &= 2 \pi \cdot \mathrm{smearing} / 2.2

        \mathrm{r_c} &= 2.2 \cdot \mathrm{smearing}

    where :math:`N` is the number of particles. The magic numbers :math:`1.3` and
    :math:`2.2` above result from tests on a CsCl system, whose unit cell is repeated 16
    times in each direction, resulting in a system of 8192 atoms.

    :param potential: the two-body potential that should be computed, implemented
        as a :class:`torchpme.potentials.Potential` object. The ``smearing`` parameter
        of the potential determines the split between real and k-space regions.
        For a :class:`torchpme.CoulombPotential` it corresponds to the
        width of the atom-centered Gaussian used to split the Coulomb potential
        into the short- and long-range parts. A reasonable value for
        most systems is to set it to ``1/5`` times the neighbor list cutoff.
    :param lr_wavelength: Spatial resolution used for the long-range (reciprocal space)
        part of the Ewald sum. More concretely, all Fourier space vectors with a
        wavelength >= this value will be kept. If not set to a global value, it will be
        set to half the smearing parameter to ensure convergence of the
        long-range part to a relative precision of 1e-5.
    :param full_neighbor_list: If set to :obj:`True`, a "full" neighbor list is
        expected as input. This means that each atom pair appears twice. If set to
        :obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.
    """

    def __init__(
        self,
        potential: Potential,
        lr_wavelength: float,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
    ):
        super().__init__(
            potential=potential,
            full_neighbor_list=full_neighbor_list,
            prefactor=prefactor,
        )
        if potential.smearing is None:
            raise ValueError(
                "Must specify range radius to use a potential with EwaldCalculator"
            )
        self.lr_wavelength: float = lr_wavelength

    def _compute_kspace(
        self,
        charges: torch.Tensor,
        cell: torch.Tensor,
        positions: torch.Tensor,
    ) -> torch.Tensor:
        # Define k-space cutoff from required real-space resolution
        k_cutoff = 2 * torch.pi / self.lr_wavelength

        # Compute number of times each basis vector of the reciprocal space can be
        # scaled until the cutoff is reached
        basis_norms = torch.linalg.norm(cell, dim=1)
        ns_float = k_cutoff * basis_norms / 2 / torch.pi
        ns = torch.ceil(ns_float).long()

        # Generate k-vectors and evaluate
        kvectors = generate_kvectors_for_ewald(ns=ns, cell=cell)
        knorm_sq = torch.sum(kvectors**2, dim=1)

        # G(k) is the Fourier transform of the Coulomb potential
        # generated by a Gaussian charge density
        # We remove the singularity at k=0 by explicitly setting its
        # value to be equal to zero. This mathematically corresponds
        # to the requirement that the net charge of the cell is zero.
        # G = 4 * torch.pi * torch.exp(-0.5 * smearing**2 * knorm_sq) / knorm_sq
        G = self.potential.lr_from_k_sq(knorm_sq)

        # Compute the energy using the explicit method that
        # follows directly from the Poisson summation formula.
        # For this, we precompute trigonometric factors for optimization, which leads
        # to N^2 rather than N^3 scaling.
        trig_args = kvectors @ (positions.T)  # [k, i]

        c = torch.cos(trig_args)  # [k, i]
        s = torch.sin(trig_args)  # [k, i]
        sc = torch.stack([c, s], dim=0)  # [2 "f", k, i]
        sc_summed_G = torch.einsum("fki,ic, k->fkc", sc, charges, G)
        energy = torch.einsum("fkc,fki->ic", sc_summed_G, sc)
        energy /= torch.abs(cell.det())

        # Remove the self-contribution: Using the Coulomb potential as an
        # example, this is the potential generated at the origin by the fictituous
        # Gaussian charge density in order to split the potential into a SR and LR part.
        # This contribution always should be subtracted since it depends on the smearing
        # parameter, which is purely a convergence parameter.
        energy -= charges * self.potential.self_contribution()

        # Step 5: The method requires that the unit cell is charge-neutral.
        # If the cell has a net charge (i.e. if sum(charges) != 0), the method
        # implicitly assumes that a homogeneous background charge of the opposite sign
        # is present to make the cell neutral. In this case, the potential has to be
        # adjusted to compensate for this.
        # An extra factor of 2 is added to compensate for the division by 2 later on
        ivolume = torch.abs(cell.det()).pow(-1)
        charge_tot = torch.sum(charges, dim=0)
        prefac = self.potential.background_correction()
        energy -= 2 * prefac * charge_tot * ivolume
        # Compensate for double counting of pairs (i,j) and (j,i)
        return energy / 2



from typing import Optional

import torch

# from ..potentials import Potential
from .kvectors import generate_kvectors_for_mesh


class KSpaceKernel(torch.nn.Module):
    r"""
    Base class defining the interface for a reciprocal-space kernel helper.

    Provides an interface to compute the reciprocal-space convolution kernel
    that is used e.g. to compute potentials using Fourier transforms. Parameters
    of the kernel in derived classes should be defined and stored in the
    ``__init__`` method.

    NB: we need this slightly convoluted way of implementing what often amounts
    to a simple, pure function of :math:`|\mathbf{k}|^2` in order to be able to
    provide a customizable filter class that can be jitted.
    """

    def __init__(self):
        super().__init__()

    def kernel_from_k_sq(self, kvectors: torch.Tensor) -> torch.Tensor:
        r"""
        Computes the reciprocal-space kernel on a grid of k points given a tensor
        containing :math:`\mathbf{k}`.

        :param kvectors: torch.tensor containing the k vector at which the
            kernel is to be evaluated.
        """
        raise NotImplementedError(
            f"kernel_from_k_sq is not implemented for '{self.__class__.__name__}'"
        )


class KSpaceFilter(torch.nn.Module):
    r"""
    Apply a reciprocal-space filter to a real-space mesh.

    The class combines the costruction of a reciprocal-space grid
    :math:`\{\mathbf{k}_n\}`
    (that should be commensurate to the grid in real space, so the class takes
    the same options as :class:`MeshInterpolator`), the calculation of
    a scalar filter function :math:`\phi(|\mathbf{k}|^2)`, defined as a function of
    the squared norm of the reciprocal space grid points, and the application
    of the filter to a real-space function :math:`f(\mathbf{x})`,
    defined on a mesh :math:`\{\mathbf{x}_n\}`.

    In practice, the application of the filter amounts to
    :math:`f\rightarrow \hat{f} \rightarrow \hat{\tilde{f}}=
    \hat{f} \phi \rightarrow \tilde{f}`

    See also the :ref:`example-kspace-demo` for a demonstration of the functionalities
    of this class.

    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param ns_mesh: toch.tensor of shape ``(3,)``
        Number of mesh points to use along each of the three axes
    :param kernel: KSpaceKernel
        A KSpaceKernel-derived class providing a ``from_k_sq`` method that
        evaluates :math:`\psi` given the square modulus of
        the k-space mesh points
    :param fft_norm: str
        The normalization applied to the forward FT. Can be
        "forward", "backward", "ortho". See :func:`torch:fft:rfftn`
    :param ifft_norm: str
        The normalization applied to the inverse FT. Can be
        "forward", "backward", "ortho". See :func:`torch:fft:irfftn`
    """

    def __init__(
        self,
        cell: torch.Tensor,
        ns_mesh: torch.Tensor,
        kernel: KSpaceKernel,
        fft_norm: str = "ortho",
        ifft_norm: str = "ortho",
    ):
        super().__init__()

        self._fft_norm = fft_norm
        self._ifft_norm = ifft_norm
        if fft_norm not in ["ortho", "forward", "backward"]:
            raise ValueError(
                f"Invalid option '{fft_norm}' for the `fft_norm` parameter."
            )
        if ifft_norm not in ["ortho", "forward", "backward"]:
            raise ValueError(
                f"Invalid option '{ifft_norm}' for the `ifft_norm` parameter."
            )

        self.kernel = kernel
        self.update(cell, ns_mesh)

    @torch.jit.export
    def update(
        self,
        cell: Optional[torch.Tensor] = None,
        ns_mesh: Optional[torch.Tensor] = None,
    ) -> None:
        """
        Update buffers and derived attributes of the instance.

        If neither ``cell`` nor ``ns_mesh`` are passed, only the filter is updated,
        typically following a change in the underlying potential. If ``cell`` and/or
        ``ns_mesh`` are given, the instance's attributes required by these will also be
        updated accordingly.

        :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th
            basis vector of the unit cell
        :param ns_mesh: toch.tensor of shape ``(3,)``
            Number of mesh points to use along each of the three axes
        """
        self._prep_kvectors(cell, ns_mesh)

        # always update the kfilter to reduce the risk it'd go out of sync if the is an
        # update in the underlaying potential
        self._kfilter = self.kernel.kernel_from_k_sq(self._k_sq)

    def forward(self, mesh_values: torch.Tensor) -> torch.Tensor:
        """
        Applies the k-space filter by Fourier transforming the given
        ``mesh_values`` tensor, multiplying the result by the filter array
        (that should have been previously computed with a call to
        :func:`update`) and Fourier-transforming back
        to real space.

        If you update the ``cell``, the ``ns_mesh`` or anything inside the ``kernel``
        object after you initlized the object, you have call :meth:`update` to update
        the object calling this method.

        .. code-block:: python

            kernel_filter.update(cell)
            kernel_filter.forward(mesh)

        :param mesh_values: torch.tensor of shape ``(n_channels, nx, ny, nz)``
            The values of the input function on a real-space mesh. Shape
            should match the shape of the filter.

        :returns: torch.tensor of shape ``(n_channels, nx, ny, nz)``
            The real-space mesh containing the transformed function values.
        """
        if mesh_values.dim() != 4:
            raise ValueError(
                "`mesh_values` needs to be a 4 dimensional tensor, got "
                f"{mesh_values.dim()}"
            )

        if mesh_values.device != self._kfilter.device:
            raise ValueError(
                "`mesh_values` and the k-space filter are on different devices, got "
                f"{mesh_values.device} and {self._kfilter.device}"
            )

        # Applying the Fourier filter involves the following substeps:
        # 1. Fourier transform the input mesh
        # 2. multiply by kernel in k-space
        # 3. transform back
        # For the Fourier transforms, we use the normalization conditions
        # that do not introduce any extra factors of 1/n_mesh.
        # This is why the forward transform (fft) is called with the
        # normalization option 'backward' (the convention in which 1/n_mesh
        # is in the backward transformation) and vice versa for the
        # inverse transform (irfft).

        dims = (1, 2, 3)  # dimensions along which to Fourier transform
        mesh_hat = torch.fft.rfftn(mesh_values, norm=self._fft_norm, dim=dims)

        if mesh_hat.shape[-3:] != self._kfilter.shape[-3:]:
            raise ValueError(
                "The real-space mesh is inconsistent with the k-space grid."
            )

        filter_hat = mesh_hat * self._kfilter

        result = torch.fft.irfftn(
            filter_hat,
            norm=self._ifft_norm,
            dim=dims,
            # NB: we must specify the size of the output
            # as for certain mesh sizes the inverse FT is not
            # well-defined
            s=mesh_values.shape[-3:],
        )

        if torch.isnan(result).any():
            raise ValueError(
                "NaNs detected in the k-space filter result. This are probably caused "
                "by an unsuitable `mesh_spacing`, resulting in a problematic grid of "
                f"shape: {list(mesh_values.shape)}. Try adjsuting the grid by using a "
                "different `mesh_spacing` value."
            )

        return result

    def _prep_kvectors(
        self, cell: Optional[torch.Tensor], ns_mesh: Optional[torch.Tensor]
    ):
        if cell is not None:
            if cell.shape != (3, 3):
                raise ValueError(
                    f"cell of shape {list(cell.shape)} should be of shape (3, 3)"
                )
            self.cell = cell

        if ns_mesh is not None:
            if ns_mesh.shape != (3,):
                raise ValueError(
                    f"shape {list(ns_mesh.shape)} of `ns_mesh` has to be (3,)"
                )
            self.ns_mesh = ns_mesh

        if self.cell.device != self.ns_mesh.device:
            raise ValueError(
                "`cell` and `ns_mesh` are on different devices, got "
                f"{self.cell.device} and {self.ns_mesh.device}"
            )

        if cell is not None or ns_mesh is not None:
            self._kvectors = generate_kvectors_for_mesh(ns=self.ns_mesh, cell=self.cell)
            self._k_sq = torch.linalg.norm(self._kvectors, dim=3) ** 2


class P3MKSpaceFilter(KSpaceFilter):
    r"""
    A specialized implementation of the k-space filter for the P3M method, with a
    cell-dependent Green's function kernel. This class does almost the same thing as
    :class:`KSpaceFilter`, but with a different, P3M-specialized filter. See `this paper
    <http://dx.doi.org/10.1063/1.477414>`_ for your reference.

    :param cell: torch.tensor of shape ``(3, 3)``, where ``cell[i]`` is the i-th basis
        vector of the unit cell
    :param ns_mesh: toch.tensor of shape ``(3,)`` Number of mesh points to use along
        each of the three axes
    :param interpolation_nodes: int The number ``n`` of nodes used in the interpolation
        per coordinate axis. The total number of interpolation nodes in 3D will be
        ``n^3``. In general, for ``n`` nodes, the interpolation will be performed by
        piecewise polynomials of degree ``n - 1`` (e.g. ``n = 4`` for cubic
        interpolation). Only the values ``1, 2, 3, 4, 5`` are supported.
    :param kernel: KSpaceKernel A KSpaceKernel-derived class providing a ``from_k_sq``
        method that evaluates :math:`\psi` given the square modulus of the k-space mesh
        points
    :param fft_norm: str The normalization applied to the forward FT. Can be "forward",
        "backward", "ortho". See :func:`torch:fft:rfftn`
    :param ifft_norm: str The normalization applied to the inverse FT. Can be "forward",
        "backward", "ortho". See :func:`torch:fft:irfftn`
    :param mode: int, 0 for the electrostatic potential, 1 for the electrostatic energy,
        2 for the dipolar torques, and 3 for the dipolar forces. For more details, see
        eq.30 of `that paper <https://doi.org/10.1063/1.3000389>`_.
    :param diff_order: int, the order of the approximation of the difference operator.
        Higher order is more accurate, but also more expensive. For more details, see
        Appendix C of `this paper <http://dx.doi.org/10.1063/1.477414>`_. The values
        ``1, 2, 3, 4, 5, 6`` are supported.
    """

    def __init__(
        self,
        cell: torch.Tensor,
        ns_mesh: torch.Tensor,
        interpolation_nodes: int,
        kernel: KSpaceKernel,
        fft_norm: str = "ortho",
        ifft_norm: str = "ortho",
        mode: int = 0,
        differential_order: int = 2,
    ):
        self.interpolation_nodes = interpolation_nodes
        if mode not in [0, 1, 2, 3]:
            raise ValueError(f"`mode` should be one of [0, 1, 2, 3], but got {mode}")
        self.mode = mode
        if differential_order not in [1, 2, 3, 4, 5, 6]:
            raise ValueError(
                f"`differential_order` should be one between 1 and 6, but got {differential_order}"
            )
        self.differential_order = differential_order

        super().__init__(cell, ns_mesh, kernel, fft_norm, ifft_norm)
        self.register_buffer(
            "_diff_coeff",
            torch.tensor(
                [  # coefficients for the difference operator
                    [1.0, 0.0, 0.0, 0.0, 0.0, 0.0],
                    [4 / 3, -1 / 3, 0.0, 0.0, 0.0, 0.0],
                    [3 / 2, -3 / 5, 1 / 10, 0.0, 0.0, 0.0],
                    [8 / 5, -4 / 5, 8 / 35, -1 / 35, 0.0, 0.0],
                    [5 / 3, -20 / 21, 5 / 14, -5 / 63, 1 / 126, 0.0],
                    [12 / 7, -15 / 14, 10 / 21, -1 / 7, 2 / 77, -1 / 465],
                ]
            ),
        )

    @torch.jit.export
    def update(
        self,
        cell: Optional[torch.Tensor] = None,
        ns_mesh: Optional[torch.Tensor] = None,
    ) -> None:
        # Cannot reuse code from `KSpaceFilter` by `super` because of `TorchScript`
        self._prep_kvectors(cell, ns_mesh)

        # always update the kfilter to ensure sync with potential
        self._kfilter = self._compute_influence(
            self._kvectors
        ) * self.kernel.kernel_from_k_sq(self._k_sq)

    def _compute_influence(self, kvectors: torch.Tensor) -> torch.Tensor:
        cell_dimensions = torch.linalg.norm(self.cell, dim=1)
        actual_mesh_spacing = (cell_dimensions / self.ns_mesh).reshape(1, 1, 1, 3)

        kh = kvectors * actual_mesh_spacing
        U2 = self._charge_assignment(kh)
        if self.mode == 0:
            # special (much simpler) case for point-charge potentials
            masked = torch.where(U2 == 0, 1.0, U2)
            return torch.where(U2 == 0, 0.0, torch.reciprocal(masked))

        D = self._differential_operator(kh, actual_mesh_spacing)
        D_to_4mode = torch.linalg.norm(D, dim=-1) ** (4 * self.mode)

        # Calculate (part of) the kernel See eq.30 of this paper
        # https://doi.org/10.1063/1.3000389 for your main reference, as well as the
        # paragraph below eq.31. Be careful that the reference force part is calculated
        # in the `KSpaceFilter` by calling `kernel_from_k_sq`.
        numerator = torch.sum(kvectors * D, dim=-1) ** self.mode
        denominator = U2 * D_to_4mode

        masked = torch.where(denominator == 0, 1.0, denominator)
        return torch.where(denominator == 0, 0.0, numerator / masked)

    def _differential_operator(
        self, kh: torch.Tensor, actual_mesh_spacing: torch.Tensor
    ) -> torch.Tensor:
        """
        The approximation to the differential operator ``ik``. The ``i`` is taken out
        and cancels with the prefactor ``-i`` of the reference force function (in our
        code, this is the kernel of `Potential`). See the Appendix C of this paper
        http://dx.doi.org/10.1063/1.477414.

        From shape (nx, ny, nz, 3) to shape (nx, ny, nz, 3)
        """
        temp = torch.zeros(kh.shape, dtype=kh.dtype, device=kh.device)
        for i, coef in enumerate(
            self._diff_coeff[self.differential_order - 1][: self.differential_order]
        ):
            temp += (coef / (i + 1)) * torch.sin(kh * (i + 1))
        return temp / (actual_mesh_spacing)

    def _charge_assignment(self, kh: torch.Tensor) -> torch.Tensor:
        """
        The Fourier transformed charge assignment function divided by the volume of one
        mesh cell, in a squared form. See eq.18 and the paragraph below eq.31 of this
        paper http://dx.doi.org/10.1063/1.477414. Be aware that the volume cancels out
        with the prefactor of the assignment function (see eq.18).

        From shape (nx, ny, nz, 3) to shape (nx, ny, nz, nd)
        """
        return torch.prod(
            torch.sinc(kh / (2 * torch.pi)),
            dim=-1,
        ) ** (self.interpolation_nodes * 2)

    update.__doc__ = KSpaceFilter.update.__doc__



import torch

from ..lib.kspace_filter import P3MKSpaceFilter
from ..lib.mesh_interpolator import MeshInterpolator
from ..potentials import Potential
from .pme import PMECalculator


class P3MCalculator(PMECalculator):
    r"""
    Potential using a particle-particle particle-mesh based Ewald (P3M).

    For getting reasonable values for the ``smearing`` of the potential class and  the
    ``mesh_spacing`` based on a given accuracy for a specific structure you should use
    :func:`torchpme.tuning.tune_p3m`. This function will also find the optimal
    ``cutoff`` for the  **neighborlist**.

    .. hint::

        For a training exercise it is recommended only run a tuning procedure with
        :func:`torchpme.tuning.tune_p3m` for the largest system in your dataset.

    :param potential: A :py:class:`Potential` object that implements the evaluation
        of short and long-range potential terms. The ``smearing`` parameter
        of the potential determines the split between real and k-space regions.
        For a :py:class:`torchpme.lib.CoulombPotential` it corresponds
        to the smearing of the atom-centered Gaussian used to split the
        Coulomb potential into the short- and long-range parts. A reasonable value for
        most systems is to set it to ``1/5`` times the neighbor list cutoff.
    :param mesh_spacing: Value that determines the umber of Fourier-space grid points
        that will be used along each axis. If set to None, it will automatically be set
        to half of ``smearing``.
    :param interpolation_nodes: The number ``n`` of nodes used in the interpolation per
        coordinate axis. The total number of interpolation nodes in 3D will be ``n^3``.
        In general, for ``n`` nodes, the interpolation will be performed by piecewise
        polynomials of degree ``n - 1`` (e.g. ``n = 4`` for cubic interpolation).
        Only the values ``1, 2, 3, 4, 5`` are supported.
    :param full_neighbor_list: If set to :py:obj:`True`, a "full" neighbor list
        is expected as input. This means that each atom pair appears twice. If
        set to :py:obj:`False`, a "half" neighbor list is expected.
    :param prefactor: electrostatics prefactor; see :ref:`prefactors` for details and
        common values.

    For an **example** on the usage for any calculator refer to :ref:`userdoc-how-to`.
    """

    def __init__(
        self,
        potential: Potential,
        mesh_spacing: float,
        interpolation_nodes: int = 4,
        full_neighbor_list: bool = False,
        prefactor: float = 1.0,
    ):
        self.mesh_spacing: float = mesh_spacing

        if interpolation_nodes not in [1, 2, 3, 4, 5]:
            raise ValueError("Only `interpolation_nodes` from 1 to 5 are allowed")
        self.interpolation_nodes: int = interpolation_nodes

        super(PMECalculator, self).__init__(
            potential=potential,
            full_neighbor_list=full_neighbor_list,
            prefactor=prefactor,
        )

        if potential.smearing is None:
            raise ValueError(
                "Must specify smearing to use a potential with P3MCalculator"
            )

        cell = torch.eye(
            3,
            device=self.potential.smearing.device,
            dtype=self.potential.smearing.dtype,
        )
        ns_mesh = torch.ones(3, dtype=int, device=cell.device)

        self.kspace_filter: P3MKSpaceFilter = P3MKSpaceFilter(
            cell=cell,
            ns_mesh=ns_mesh,
            interpolation_nodes=self.interpolation_nodes,
            kernel=self.potential,
            mode=0,  # Green's function for point-charge potentials
            differential_order=2,  # Order of the discretization of the differential operator
            fft_norm="backward",
            ifft_norm="forward",
        )

        self.mesh_interpolator: MeshInterpolator = MeshInterpolator(
            cell=cell,
            ns_mesh=ns_mesh,
            interpolation_nodes=self.interpolation_nodes,
            method="P3M",
        )
